# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
'''
    The model (table) module

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 04-28-2023 06:59:46
    `name`: model
    * @xxx [04-28-2023 06:59:59]: documentation for model
'''






from dataclasses import dataclass
from typing import Iterable, Union
from datetime import datetime
from datetime import timezone


import colemen_utils as c


import volent.settings.types as _t
import volent.settings as _settings
from volent.Column import Column as _column
from volent.Relationship import Relationship as _relationship
from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint
from volent.FullTextIndex import FullTextIndex as _fullTextIndex
from volent.mixins import OrderedClass,MySQLGeneratorMixin
from volent.query.Insert import Insert as _insert
from volent.query.Select import Select as _select
from volent.query.Update import Update as _update
from volent.query.Delete import Delete as _delete

@dataclass
# class Model(MySQLGeneratorMixin):
class Model(MySQLGeneratorMixin,metaclass=OrderedClass):
    _is_root= False
    _model_name:str = None
    '''The name of this models table'''
    _database_name:str = None
    '''The name of the database that this model belongs to.'''
    _database:_t.database_type = None
    '''A reference to the database instance.'''
    _description:str = None
    '''A description of this model that is applied to the database table.'''
    _columns:Iterable[_t.column_type] = None
    '''A list of column instances'''
    _column_lookup:dict[str,_t.column_type] = None
    '''A list of primary column instances.'''
    _relationships:Iterable[_t.relationship_type] = None
    '''A list of relationship instances'''
    _unique_constraints:Iterable[_t.unique_constraint_type] = None
    _unique_constraint_lookup:dict[str,_t.unique_constraint_type] = None
    '''A list of full unique constraint instances'''
    _full_text_indexes:Iterable[_t.full_text_index_type] = None
    '''A list of full text index instances'''
    primary_column:_t.column_type = None
    _primary_columns:list[_t.column_type] = None

    __unique_prop_keys = None
    _saved:bool = False

    _parent_models:Iterable[_t.model_type] = None
    '''A list of models that are a parent to this model.'''
    _child_models:Iterable[_t.model_type] = None
    '''A list of models that children of this model.'''

    _longest_column_name:tuple[int,str] = None
    _relationships_updated:bool = False
    '''True if all relationships have had their parent models assigned.'''

    _gen_meta_columns:bool = None
    # _should_create:bool = False
    # _existing_table_data:dict = None

    def __init__(self,_is_root=False,**kwargs) -> None:
        '''
            Create a new model instance.

            ----------

            Arguments
            -------------------------
            [`_is_root`=False] {bool}
                This is used internally to determine if this model instance is the first of its kind.
                You really don't need to worry about it.

            Keyword Arguments
            -------------------------
            The keyword arguments are used to apply values to the columns being created for this model.

            So you can supply a spread dictionary with keys that match column names, and the values will
            be automatically applied to the columns.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:00:11
            `memberOf`: Model
            `version`: 1.0
            `method_name`: model
            * @xxx [04-28-2023 07:02:43]: documentation for model
        '''
        if _is_root is True:
            # _new_get_cls_vars(self)
            _settings.globe.Volent.register(self)
        else:
            self._is_root = False
            # _settings.globe.Volent.register(self)
            # print(f"self.name:{self.name}")
            # mdl = _settings.globe.Volent.get_model(self.name)
            # self.database = mdl.database


        for k,v in kwargs.items():
            col = self.get_column(k)
            if col is not None:
                if col.is_primary is True and v is None:
                    continue
                if v != _t.no_default:
                    col.value = v
        self._parent_models = []
        self._child_models = []
        self.__gen_auto_columns()

    def __repr__(self) -> str:
        return f"<Model: {self._model_name} | Columns:{len(self.columns)}>"

    def _cpid(self,data,locals):

        d = {}
        for x in data.args:
            # if x in ["self","_is_root"]:
            #     continue
            d[x] = locals[x]
            # print(f"x : {x}")
        d = self.filter_dict_by_columns(d)
        # for k,v in d.items():
        #     col = self.get_column(k)
        #     if col is not None:
        #         print(f"value for column:{col.name} found: {v}")

        return d

    def _get_attrs_from_parent(self):
        '''Used internally to apply the root model attributes to this instance.

        Essentially this just copies the database reference to all child instances of a model.
        '''
        if self._is_root is False:
            # print(f"self.name:{self.name}")
            mdl = _settings.globe.Volent.get_model(self.model_name)
            self._database = mdl.database

    def __setattr__(self, name, value):
        # This is useful for setting the value of a column without accessing the column itself
        # Not critically necessary but it is convenient.
        try:
            # @Mstep [] Attempt to retrieve a column with a matching name.
            nv = super().__getattribute__(name)
            # @Mstep [IF] if the attribute exists and its a column instance.
            if isinstance(nv,(_column)):
                # @Mstep [] set the column value to the value provided.
                nv.value = value
                # @Mstep [RETURN] 
                return
        except AttributeError:
            pass


        super().__setattr__(name, value)

        if not name in self._order:
            self._order.append(name)
            if isinstance(value,_column):
                cols = [] if self._columns is None else self._columns
                col_lookup = {} if self._column_lookup is None else self._column_lookup
                primes = [] if self._primary_columns is None else self._primary_columns
                if value.name is None:
                    value.name = name
                value.model = self
                cols.append(value)
                if value.is_primary is True:
                    self.primary_column = value
                    primes.append(value)
                _=value.unique
                _=value.fulltext
                col_lookup[name] = value
                self._primary_columns = primes
                self._column_lookup = col_lookup
                self._columns = cols
                self._longest_column_name = c.arr.longest_string(list(col_lookup.keys()))

            self._unique_constraints = [] if self._unique_constraints is None else self._unique_constraints
            self._unique_constraint_lookup = {} if self._unique_constraint_lookup is None else self._unique_constraint_lookup
            if isinstance(value,_uniqueConstraint):
                # print(f"registering unique constraint")

                value.model = self
                if value.name is None:
                    value.name = name
                skip = False
                for uq in self._unique_constraints:
                    if uq == value:
                        # print(f"Duplicate unique constraint found: {name} and {uq.name}")
                        skip = True
                if skip is False:
                    self._unique_constraint_lookup[name] = value
                    self._unique_constraints.append(value)

            self._full_text_indexes = [] if self._full_text_indexes is None else self._full_text_indexes
            if isinstance(value,_fullTextIndex):

                value.model = self
                if value.name is None:
                    value.name = name

                skip = False
                for uq in self._full_text_indexes:
                    if uq == value:
                        # print(f"Duplicate fulltext index found: {name} and {uq.name}")
                        skip = True
                if skip is False:
                    self._full_text_indexes.append(value)


            self._relationships = [] if self._relationships is None else self._relationships
            if isinstance(value,_relationship):

                value.child_model = self
                if value.name is None:
                    value.name = name
                self._relationships.append(value)


        return value

    def ordered_attrs(self, with_order=False):
        return [(k,getattr(self, k)) for k in self._order if k != '_order' or with_order]

    @property
    def summary(self):
        '''
            Get this Model's summary

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 08:25:40
            `@memberOf`: Model
            `@property`: summary
        '''
        value = {
            "name":self.model_name,
            "description":self.model_description,
            "columns":[x.summary for x in self.columns],
            "relationships":[x.summary for x in self.relationships],
            "unique_constraints":[x.summary for x in self.unique_constraints],
            "fulltext_indexes":[x.summary for x in self.full_text_indexes],
            "parent_tables":[x.model_name for x in self._parent_models],
            "child_tables":[x.model_name for x in self._child_models],
        }


        return value

    @property
    def model_name(self)->str:
        '''
            Get this Model's name

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 14:15:56
            `@memberOf`: Model
            `@property`: name
        '''
        value = self._model_name
        if value is None:
            keys = ["__table_name__","__tablename__"]
            for key in keys:
                if hasattr(self,key):
                    value = getattr(self,key)
                    break
            # if hasattr(self,"__table_name__"):
                # value = getattr(self,"__table_name__")
            self._model_name = value
        return value

    @property
    def database_name(self)->str:
        '''
            Get this Model's database_name

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 15:06:52
            `@memberOf`: Model
            `@property`: database_name
        '''
        value = self._database_name
        if value is None:
            keys = ["__database_name__","__databasename__"]
            for key in keys:
                if hasattr(self,key):
                    value = getattr(self,key)
                    break
            # if hasattr(self,"__database_name__"):
                # value = getattr(self,"__database_name__")
            self._database_name = value
        return value

    @property
    def database(self)->_t.database_type:
        '''
            Get this Model's database if it has been associated to one.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 09:48:39
            `@memberOf`: Model
            `@property`: database
        '''

        value = self._database
        if value is None:
            self._get_attrs_from_parent()
            value = self._database
        return value

    @property
    def model_description(self)->str:
        '''
            Get this Model's description

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 15:06:52
            `@memberOf`: Model
            `@property`: description
        '''
        value = self._description
        if value is None:
            if hasattr(self,"__description__"):
                value = getattr(self,"__description__")
            else:
                value = self.__class__.__doc__
            self._description = value
        return value

    @property
    def data(self)->dict:
        '''
            Retrieve a dictionary of columns and their values.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 14:48:54
            `@memberOf`: Model
            `@property`: data
        '''
        self.__gen_auto_columns()
        value = {}
        for col in self.columns:
            value[col.name] = col.value
        return value


    @property
    def gen_meta_columns(self):
        '''
            Get this Model's gen_meta_columns setting
            
            if True, the timestamp,deleted and modified_timestamp columns are automatically created on 
            this model.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-28-2023 10:57:56
            `@memberOf`: Model
            `@property`: gen_meta_columns
        '''
        value = self._gen_meta_columns
        if value is None:
            val = False
            keys = ["__gen_meta_columns__","__meta_columns__"]
            for key in keys:
                if hasattr(self,key):
                    val = getattr(self,key)
                    break
            # if hasattr(self,"__table_name__"):
                # value = getattr(self,"__table_name__")
            self._gen_meta_columns = val
            value = val
        return value

    def __gen_auto_columns(self):
        '''
            If the __meta_columns__ key is True, this will create the
            timestamp,deleted and modified_timestamp columns on this model automatically.

            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 11:07:25
            `memberOf`: Model
            `version`: 1.0
            `method_name`: __gen_auto_columns
            * @xxx [04-28-2023 11:08:31]: documentation for __gen_auto_columns
        '''
        if self.gen_meta_columns is False:
            return


        from volent.data_types import Integer
        cut = round(datetime.now(tz=timezone.utc).timestamp())

        ct = _column(Integer(), nullable=True, on_insert=cut, comment='The unix timestamp of when this was created.')
        dt = _column(Integer(), nullable=True, default=None, comment='The unix timestamp of when this was deleted, null otherwise.')
        mt = _column(Integer(), nullable=True, default=None, on_update=cut,on_insert=cut, comment='The unix timestamp of when this was last modified, null otherwise.')
        setattr(self,"timestamp",ct)
        setattr(self,"modified_timestamp",mt)
        setattr(self,"deleted",dt)

        # self._columns.append(ct)
        # self._columns.append(mt)
        # self._columns.append(dt)

    def __gen_col_sums(self):
        primes = []
        col_lookup = {}

        for col in self.columns:
            col_lookup[col.name] = col
            if col.is_primary is True:
                primes.append(col)

        self._primary_columns = primes
        self._column_lookup = col_lookup
        self._longest_column_name = c.arr.longest_string(list(col_lookup.keys()))

    def get_column(self,name:str)->_t.column_type:
        '''
            Retrieve a column by searching its name.

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the column to search for.

            Return {column,None}
            ----------------------
            THe column instance if successful, None otherwise

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 09:46:53
            `memberOf`: Model
            `version`: 1.0
            `method_name`: get_column
            * @xxx [04-28-2023 09:47:36]: documentation for get_column
        '''
        value = None
        if self._column_lookup is None:
            # print(f"GENERATING COLUMN LOOKUP {'-'*80}")
            self.__gen_col_sums()


        if name in self._column_lookup:
            # print(f"self.name: {self.name}")
            return self._column_lookup[name]

        for col in self.columns:
            if col.name == name:
                return col

        return None

    def filter_dict_by_columns(self,data:dict):
        out = {}
        for col in self.columns:
            if col.name in data:
                out[col.name] = data[col.name]
        return out


    @property
    def columns(self)->Iterable[_t.column_type]:
        '''
            Get this Model's columns

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 14:07:35
            `@memberOf`: Model
            `@property`: columns
        '''
        value = self._columns
        self.__gen_auto_columns()
        if value is None:
            # print(f"GENERATING COLUMN LIST {'-'*80}")
            dif = self._unique_prop_keys
            # df_props = dir(Model(_is_root=False))
            # # # @Mstep [] gather the props of this instance.
            # props = dir(self)
            # # # @Mstep [] find the props that exist on this instance and not on the base.
            # dif = c.arr.find_list_diff(props,df_props)
            # dif = dir(self)
            value = []
            # color = c.rand.option(["magenta","yellow","green","cyan"])
            for prop in dif:
                name = prop
                # if isinstance(val,(dict)):
                    # props[name] = val
                val = getattr(self,prop)
                if isinstance(val,(_column)):
                    # c.con.log(f"located Column: {name}","green")
                    val.name = name
                    val.model = self
                    if val.is_primary is True:
                        self.primary_column = val
                    value.append(val)

        return value


        order_cols = []
        for k in list(self.ordered_attrs(True)):
            for col in value:
                if col.name == k[0]:
                    order_cols.append(col)
                    break
        value = order_cols
            # order_cols = []
            # for k in list(self.__fields__):
            #     for col in value:
            #         if col.name == k:
            #             order_cols.append(col)
            #             break
            # value = order_cols



            # return props
        self._columns = value
        return value

    @property
    def primary_columns(self)->Iterable[_t.column_type]:
        '''
            Get this Model's primary_columns

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-24-2023 09:17:42
            `@memberOf`: Model
            `@property`: primary_columns
        '''
        if self._primary_columns is None:
            # print(f"GENERATING PRIMARY COLUMNS {'-'*80}")
            self.__gen_col_sums()
        # print(f"self._primary_columns: {self._primary_columns}")
        return self._primary_columns
        value = []
        # print(f"self.columns: {self.columns}")
        for col in self.columns:
            # c.con.log(f"col.name:{col.name}")
            if col.is_primary is True:
                # print(f"{col.name}.is_primary: {col.is_primary} {type(col.is_primary)}")
                value.append(col)
        return value

    def get_unique_constraint(self,name:str)->_t.unique_constraint_type:
        '''
            Retrieve a unique constraint by its name.

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the unique constraint to search for.

            Return {UniqueConstraint}
            ----------------------
            The unique constraint instance if successful, None otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 08:29:07
            `memberOf`: Model
            `version`: 1.0
            `method_name`: get_unique_constraint
            * @xxx [04-28-2023 08:30:04]: documentation for get_unique_constraint
        '''
        if name in self._unique_constraint_lookup:
            return self._unique_constraint_lookup[name]
        for u in self.unique_constraints:
            if u.name == name:
                return u
        return None

    @property
    def unique_constraints(self)->Iterable[_t.unique_constraint_type]:
        '''
            Get a list of unique constraint instances from this model.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-25-2023 08:24:42
            `@memberOf`: Model
            `@property`: unique_constraints
        '''
        # return None
        value = self._unique_constraints

        if value is None:
            # print(f"GENERATING UNIQUE_CONSTRAINTS LIST {self.name} {'-'*80}")
            dif = self._unique_prop_keys
            # dif = dir(self)
            value = []
            # color = c.rand.option(["magenta","yellow","green","cyan"])
            for prop in dif:
                name = prop
                val = getattr(self,prop)
                if isinstance(val,(_uniqueConstraint)):
                    c.con.log(f"located unique constraint","magenta")
                    val.name = name
                    val.model = self
                    value.append(val)

            self._unique_constraints = value
        return value

    @property
    def full_text_indexes(self)->Iterable[_t.full_text_index_type]:
        '''
            Get a list of fulltext index instances from this model.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-28-2023 06:49:39
            `@memberOf`: Model
            `@property`: full_text_indexes
        '''
        value = self._full_text_indexes

        if value is None:
            # print(f"GENERATING FULL_TEXT_INDEX LIST {'-'*80}")
            dif = self._unique_prop_keys
            value = []
            for prop in dif:
                name = prop
                val = getattr(self,prop)
                if isinstance(val,(_fullTextIndex)):
                    c.con.log(f"located full text index","magenta")
                    val.name = name
                    val.model = self
                    value.append(val)

            self._full_text_indexes = value
        return value

    @property
    def relationships(self)->Iterable[_t.relationship_type]:
        '''
            Get a list of relationships instances associated to this table.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-11-2023 13:47:00
            `@memberOf`: Model
            `@property`: columns
        '''

        value = self._relationships

        if self._relationships_updated is False:
            # @Mstep [LOOP] iterate all relationship instances
            for val in value:
                # @Mstep [] assign the parent_model and parent_column to the relationship.
                # val.name = name
                val.child_model = self
                val.parent_model = _settings.globe.Volent.get_model(val.parent)
                val.parent_column = _settings.globe.Volent.get_column(val.parent)

                # value.append(val)
                val.parent_model.add_child_model(self)
                self.add_parent_model(val.parent_model)
            self._relationships_updated = True


        if value is None:
            # This should pretty much never execute, the registration of relationships is handled in __setattr__
            c.con.log(f"Model.relationships","magenta")
            dif = self._unique_prop_keys

            value = []
            for prop in dif:
                name = prop
                val:_t.relationship_type = getattr(self,prop)
                if isinstance(val,(_relationship)):
                    val.name = name
                    val.child_model = self
                    val.parent_model = _settings.globe.Volent.get_model(val.parent)
                    val.parent_column = _settings.globe.Volent.get_column(val.parent)

                    value.append(val)
                    val.parent_model.add_child_model(self)
                    self.add_parent_model(val.parent_model)

            self._relationships = value
        return value

    @property
    def _unique_prop_keys(self):
        '''
            Get this Model's _unique_prop_keys

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 15:45:07
            `@memberOf`: Model
            `@property`: _unique_prop_keys
        '''
        value = self.__unique_prop_keys
        if value is None:
            df_props = dir(Model(_is_root=False))
            # # @Mstep [] gather the props of this instance.
            props = dir(self)
            # # @Mstep [] find the props that exist on this instance and not on the base.
            value = c.arr.find_list_diff(props,df_props)
            self.__unique_prop_keys = value

        return value

    @property
    def longest_column_name(self)->tuple:
        '''
            Get this Model's longest_column_name

            A tuple containing the longest column name for this table and how many characters it has.

            (19,"kitties and titties")

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-24-2023 05:53:18
            `@memberOf`: Model
            `@property`: longest_column_name
        '''
        if self._longest_column_name is None:
            self.__gen_col_sums()
        return self._longest_column_name
        cnames = []
        for col in self.columns:
            cnames.append(col.name)
        return c.arr.longest_string(cnames)

    def volent_register_subs(self):
        '''Have this model register its child entities

        - Columns
        - Unique Constraints
        - Relationships
        '''

        for uc in self.unique_constraints:
            _settings.globe.Volent.register(uc)

        for rel in self.relationships:
            _settings.globe.Volent.register(rel)

        for col in self.columns:
            _settings.globe.Volent.register(col)



    # ---------------------------------------------------------------------------- #
    #                                 QUERY METHODS                                #
    # ---------------------------------------------------------------------------- #

    def insert(self,data:Union[dict,_t.schema_type]=None)->_insert:
        '''
            Create an insert query on this model.

            ----------

            Arguments
            -------------------------
            `data` {dict,schema}
                The data to be inserted.

            Return {insert}
            ----------------------
            An insert query instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 09:48:39
            `memberOf`: Model
            `version`: 1.0
            `method_name`: insert
            * @xxx [04-28-2023 09:49:25]: documentation for insert
        '''
        # self.database.run('show variables like "max_connections";')
        # print(self.database.fetchall())
        if isinstance(data,(dict)) is False:
            if data is None:
                data = self.data
            from volent.Schema import Schema as _schema
            if isinstance(data,_schema):
                data.model = self
                data = data.dump(self)
            

        ins = _insert(self,data)
        # ins.database = self.database
        # result = ins.execute()
        return ins

    def select(self,*columns)->_select:
        '''
            Create a select query on this model.

            ----------

            Arguments
            -------------------------
            [`*columns`] {}
                The columns to select if None are provided, all columns are selected.

            Return {select}
            ----------------------
            A select query instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 09:49:36
            `memberOf`: Model
            `version`: 1.0
            `method_name`: select
            * @xxx [04-28-2023 09:51:05]: documentation for select
        '''
        s = _select(self,columns=columns)
        return s

    def update(self,*columns)->_t.update_query_type:
        '''
            Create an update query on this model.

            ----------

            Keyword Arguments
            -------------------------
            [`*columns`] {}
                The columns to update and their values.

            Return {update}
            ----------------------
            An update query instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 09:49:36
            `memberOf`: Model
            `version`: 1.0
            `method_name`: update
            * @xxx [04-28-2023 09:51:05]: documentation for update
        '''
        print(f"columns: {columns}")
        s = _update(self,columns=columns)
        return s

    def delete(self)->_delete:
        '''
            Create a delete query on this model.

            ----------

            Return {delete}
            ----------------------
            An delete query instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 09:49:36
            `memberOf`: Model
            `version`: 1.0
            `method_name`: delete
            * @xxx [04-28-2023 09:51:05]: documentation for delete
        '''
        # print(f"columns: {columns}")
        s = _delete(self)
        return s

    def add_parent_model(self,model:_t.model_type):
        '''
            Add a parent model to this model.

            This is for internal use do not arbitrarily add models here, they wont do anything.
            ----------

            Arguments
            -------------------------
            `model` {model}
                The parent model to add.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-21-2023 08:25:10
            `memberOf`: Model
            `version`: 1.0
            `method_name`: add_parent_model
            * @xxx [04-21-2023 08:26:12]: documentation for add_parent_model
        '''
        if isinstance(model,Model) is False:
            raise TypeError(f"Expects Model type, received {type(model)}")
        self._parent_models.append(model)

    def add_child_model(self,model:_t.model_type):
        '''
            Add a child model to this model.

            This is for internal use do not arbitrarily add models here, they wont do anything.
            ----------

            Arguments
            -------------------------
            `model` {model}
                The child model to add.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-21-2023 08:25:10
            `memberOf`: Model
            `version`: 1.0
            `method_name`: add_child_model
            * @xxx [04-21-2023 08:26:12]: documentation for add_child_model
        '''
        if isinstance(model,Model) is False:
            raise TypeError(f"Expects Model type, received {type(model)}")
        self._child_models.append(model)



# def get_truth(inp, relate, cut):
#     ops = {'>': operator.gt,
#         '<': operator.lt,
#         '>=': operator.ge,
#         '<=': operator.le,
#         '==': operator.eq}
#     return ops[relate](inp, cut)

#     def insert(self,data)->tuple:
#         if isinstance(data,(dict)) is False:
#             raise ValueError("Data should be a dictionary.")

#         data = self.filter_dict_by_columns(data)
#         column_list = ', '.join(list(data.keys()))
#         # total_keys = len(list(data.keys()))
#         ph = []
#         for k in list(data.keys()):
#             ph.append("%s")
#         placeholders = ', '.join(ph)
#         template = f"""INSERT INTO {self.quoted_name} ({column_list}) VALUES ({placeholders})"""
#         data_tuple = tuple(list(data.values()))
#         # print(f"template: {template}")
#         # print(f"data_tuple: {data_tuple}")
#         return (template,data_tuple)



#     def filter_dict_by_columns(self,data:dict)->dict:
#         output = {}
#         for k,v in data.items():
#             col = self.get_column(k)
#             if col is not None:
#                 if v is None and col.nullable is False:
#                     continue
#                 output[k] = v
#         return output

# # def _new_get_cls_vars(instance):


#     # df_props = dir(Model(_is_root=False))
#     # # # @Mstep [] gather the props of this instance.
#     # props = dir(instance)
#     # # # @Mstep [] find the props that exist on this instance and not on the base.
#     # dif = c.arr.find_list_diff(props,df_props)
#     dif = dir(instance)
#     value = []
#     # color = c.rand.option(["magenta","yellow","green","cyan"])
#     # if 'title' in instance.__dict__:
#         # print(f"instance.__dict__['title']: {instance.__dict__['title']}")
#     for prop in dif:
#         name = prop
#         # if isinstance(val,(dict)):
#             # props[name] = val
#         val = getattr(instance,prop)
#         if isinstance(val,(_column,_uniqueConstraint,_relationship)):
#             # print(f"prop: {prop}")
#             # if name in instance.__dict__:
#                 # c.con.log(f"adding {name} to instance dict")
#                 # instance.__dict__[name] = value
#             # print(f"instance.__dict__: {instance.__dict__}")

#             setattr(instance,name,val)
#             # instance.set_instance_var(instance,name,val)
#             # val.name = name
#             # val.model = instance
#             # if val.is_primary is True:
#             #     self.primary_column = val
#             # value.append(val)

#     # order_cols = []
#     # for k in list(self.__fields__):
#     #     for col in value:
#     #         if col.name == k:
#     #             order_cols.append(col)
#     #             break
#     # value = order_cols



#         # return props
#         # self._columns = value
#     return instance











"""API tests"""
# pylint: disable=invalid-name

from dataclasses import dataclass
from datetime import datetime
from datetime import timezone


import colemen_utils as c
import volent

# import volent.Volent as volent
from volent.Volent import Schema,Field,validators as _v
from volent.data_types import String,Integer,EncodedPrimary




@dataclass
class SuccessfulSchema(Schema):
    '''The publicly visible data for a task'''


    def __init__(
        self,
        ) -> None:
        # self.task_id = Field(required=True,nullable=False)
        # self.target_id = Field(data_type=EncodedPrimary())
        self.limit = Field(data_type=Integer(),default=50,validate=[_v.OneOf([25,50,100])],description="The number of results to return.")
        self.offset = Field(data_type=Integer(),default=0,validate=[_v.Range(0)])
        self.start_timestamp = Field(data_type=Integer(),default=None)
        self.end_timestamp = Field(data_type=Integer(),default=None)

    def valid_timestamps(self):
        # print(f"validating schema timestamps")
        s = self.start_timestamp.value
        e = self.end_timestamp.value

        if s is not None and e is not None:
            if e > s:
                raise ValueError("Start timestamp must be after the end timestamp")







class TestSchemaDict:
    """TEST Schema Validation of dictionaries"""

    def test_successful_schema(self):
        """Confirm that we can create and validate a schema instance."""
        # assert True

        v = volent.Volent()

        data = {
            # "target_id":c.string.string_encode_int(50),
            # "target_id":123456,
            "limit":50,
            "offset":50,
            # "startTimestamp":round(datetime.now(tz=timezone.utc).timestamp()) - (86400 * 8),
            "startTimestamp":500,
            "endTimestamp":50,
        }
        ts = SuccessfulSchema()
        dumped = ts.dump(data)
        assert isinstance(dumped,(dict)) is True

    def test_failed_schema(self):
        """Confirm that we can create and validate a schema instance."""
        # assert True

        v = volent.Volent()

        data = {
            "limit":50,
            "offset":-50,
            "startTimestamp":500,
            "endTimestamp":50,
        }
        try:
            ts = SuccessfulSchema()
            dumped = ts.dump(data)
        except Exception as e:
            assert True
            return
        assert False

    def test_custom_validator(self):
        """Confirm that custom validation methods are successfully called"""
        # assert True

        v = volent.Volent()

        data = {
            # "target_id":c.string.string_encode_int(50),
            # "target_id":123456,
            "limit":50,
            "offset":50,
            # "startTimestamp":round(datetime.now(tz=timezone.utc).timestamp()) - (86400 * 8),
            "startTimestamp":50,
            "endTimestamp":500,
        }
        try:
            ts = SuccessfulSchema()
            dumped = ts.dump(data)
        except Exception as e:
            assert True
            return
        assert False

    def test_open_api_data(self):
        """Confirm that the schema produces the correct open api data dictionary"""
        # assert True

        v = volent.Volent()

        data = {
            "limit":50,
            "offset":50,
            "startTimestamp":500,
            "endTimestamp":50,
        }
        ts = SuccessfulSchema()
        dumped = ts.dump(data)
        oa = ts.open_api_data('path')
        correct = [{'name': 'limit', 'description': 'The number of results to return.', 'in': 'path', 'type': 'integer', 'required': False, 'default': 50}, {'name': 'offset', 'description': None, 'in': 'path', 'type': 'integer', 'required': False, 'default': 0}, {'name': 'start_timestamp', 'description': None, 'in': 'path', 'type': 'integer', 'required': False, 'default': None}, {'name': 'end_timestamp', 'description': None, 'in': 'path', 'type': 'integer', 'required': False, 'default': None}]
        if oa == correct:
            return

        assert False



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

# from dataclasses import dataclass
# from typing import Iterable, OrderedDict, Union
# from collections import OrderedDict
# from abc import ABC, abstractmethod
import re

import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator

class Equal(Validator):

    field_name = None
    default_message = "Must be Equal to {comparable}"


    def __init__(
        self,
        comparable
        ):
        self.comparable = comparable
        # if equal is not None and any([min, max]):
        #     raise ValueError(
        #         "The `equal` parameter was provided, maximum or "
        #         "minimum parameter must not be provided."
        #     )

        # self.min = min
        # self.max = max
        # self.equal = equal

    def _format_error(self,value) -> str:
        if self.field_name is not None:
            return  f"{self.field_name} must match {self.comparable}"
        return self.default_message.format(input=value, comparable=self.comparable)

    def _repr_args(self) -> str:
        return f"comparable={self.comparable!r}"

    def __call__(self, value: str,field_name:str=None) -> str:
        self.field_name = field_name
        if value != self.comparable:
            raise ValidationError(self._format_error(value))

        return value


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable
import colemen_utils as c
import volent.settings as _settings
import volent.settings.types as _t
from volent.data_types.TypeBase import TypeBase as _type_base

@dataclass
class Decimal(_type_base):


    def __init__(self,data_length:int=10,decimal:int=0) -> None:
        super().__init__(data_length)
        self.min_data_length = 0
        self.max_data_length = 65
        self.sql_type_name = "Decimal"
        self.python_data_type = (float)
        self.open_api_data_type = "float"

        self._validate_data_length(data_length,_settings.control.varchar_default_length)
        if decimal > 30 or decimal < 0:
            raise ValueError("The decimal for a Decimal must range from 0 to 10")

        self.decimal = decimal


    @property
    def sql(self):
        '''
            Get this Decimals SQL data type

            Decimal(10,8)

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-24-2023 04:23:20
            `@memberOf`: String
            `@property`: mysql
        '''
        if _settings.globe.flavor == "mysql":
            value = f"{self.flavored_name}({self.data_length},{self.decimal})"
        if _settings.globe.flavor == "sqlite":
            value = f"{self.flavored_name}({self.data_length},{self.decimal})"
        return value



    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} : {self.data_length} {self.decimal}>"


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
'''
    The main module of the volent library.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 04-28-2023 07:03:18
    `name`: volent
    * @xxx [04-28-2023 07:03:34]: documentation for volent
'''



from dataclasses import dataclass
import os
from string import Template
from typing import Iterable, Union
import colemen_utils as c
import volent.settings as _settings



@dataclass
class ReverseEngineer():
    def __init__(
        self
        ):
        pass

    def from_dict(self,data:Union[str,dict]):
        if isinstance(data,(str)):
            if c.file.exists(data):
                data = c.file.read.as_json(data)
        # from volent.Model import Model as Model
        dbs = data['databases']
        for db in dbs:
            models = []
            for model in db['models']:
                models.append(_gen_model_from_dict(model))


MODEL_CREATE_TEMPLATE = """
@dataclass
class $class_name(Model):
    '''$description'''
    __table_name__ = "$name"
    __database_name__ = "$data_base_name"
    __meta_columns__ = $meta_columns

    def __init__(
        self,
$col_args
        _is_root:bool = False
        ) -> None:
        data = {}

$col_declarations
$relationship_declarations
$unique_constraint_declarations
$fulltext_declarations
        
        
        for col in self.columns:
            lcl = locals()
            if col.name in lcl:
                data[col.name] = lcl[col.name]


        super().__init__(_is_root=_is_root,**data)

"""
def _gen_model_from_dict(data:dict):

    def __col_arg_list(cols:list):
        args = []
        for col in cols:
            val = f"{' ' * 8}{col['name']}:{col['data_type']['python_type_name']} = None"
            args.append(val)
        return ',\n'.join(args)

    def __col_declarations(cols:list):
        args = []
        for col in cols:
            args.append(_gen_column_create(col))
        return '\n'.join(args)
    def __relationship_declarations(rels:list):
        args = []
        for rel in rels:
            args.append(_gen_relationship(rel))
        return '\n'.join(args)
    def __unique_constraint_declarations(rels:list):
        args = []
        for rel in rels:
            args.append(_gen_unique_constraint(rel))
        return '\n'.join(args)
    def __fulltext_declarations(rels:list):
        args = []
        for rel in rels:
            args.append(_gen_fulltext_index(rel))
        return '\n'.join(args)


    s = Template(MODEL_CREATE_TEMPLATE)
    name = c.obj.get_kwarg('name',None,(str),**data)
    class_name = c.obj.get_kwarg('class_name',c.string.to_pascal_case(name),(str),**data)
    description = c.obj.get_kwarg('description',f"Defines a {name} entity",(str),**data)
    data_base_name = c.obj.get_kwarg(['data_base_name','database_name'],"",(str),**data)
    meta_columns = c.obj.get_kwarg('meta_columns',True,(bool),**data)



    val = s.substitute(
        name=name,
        class_name=class_name,
        description=description,
        data_base_name=data_base_name,
        meta_columns=meta_columns,
        col_args=__col_arg_list(data['columns']),
        col_declarations=__col_declarations(data['columns']),
        relationship_declarations=__relationship_declarations(data['relationships']),
        unique_constraint_declarations=__unique_constraint_declarations(data['unique_constraints']),
        fulltext_declarations=__fulltext_declarations(data['fulltext_indexes']),

    )

    return val

COLUMN_CREATE_TEMPLATE = """${indent}self.$name = Column($data_type($data_length)$args)"""
def _gen_column_create(col:dict):
    indent = " " * 8
    s = Template(COLUMN_CREATE_TEMPLATE)

    unique = c.obj.get_kwarg('unique',None,(bool),**col)
    fulltext = c.obj.get_kwarg('fulltext',None,(bool),**col)
    is_foreign_key = c.obj.get_kwarg('is_foreign_key',None,(bool),**col)
    is_primary = c.obj.get_kwarg('is_primary',None,(bool),**col)
    auto_increment = c.obj.get_kwarg('auto_increment',None,(bool),**col)
    nullable = c.obj.get_kwarg('nullable',None,(bool),**col)
    on_insert = c.obj.get_kwarg('on_insert',"no_on_insert",None,**col)
    on_update = c.obj.get_kwarg('on_update',"no_on_update",None,**col)
    default = c.obj.get_kwarg('default',None,None,**col)
    comment = c.obj.get_kwarg('comment',None,(str),**col)

    args = []

    if unique is True:
        args.append("unique=True")

    if fulltext is True:
        args.append("fulltext=True")

    if is_foreign_key is True:
        args.append("is_foreign_key=True")

    if is_primary is True:
        args.append("is_primary=True")

    if auto_increment is True:
        args.append("auto_increment=True")

    if nullable is True:
        args.append("nullable=True")

    if on_insert != "no_on_insert":
        args.append(f"on_insert={col['on_insert']}")

    if on_update != "no_on_update":
        args.append(f"on_update={col['on_update']}")

    if default != "no default":
        args.append(f"default={col['default']}")

    if isinstance(comment,(str)):
        if len(comment) > 0:
            comment = comment.replace("'","\'")
            args.append(f"comment='{comment}'")

    arg_string_list = ""
    if len(args) > 0:
        arg_string_list = ', '.join(args)
        arg_string_list = f",{arg_string_list}"

    data_length = ""
    if isinstance(col['data_type']['data_length'],(int)):
        data_length = col['data_type']['data_length']


    val = s.substitute(
        indent=indent,
        name=col['name'],
        data_type=col['data_type']['class_name'],
        data_length=data_length,
        args=arg_string_list,
    )


    return val

RELATIONSHIP_CREATE_TEMPLATE = """${indent}self.$name = Relationship(self.$child_column,"$parent_string",on_delete="CASCADE",on_update="CASCADE")"""
def _gen_relationship(rel:dict):
    indent = " " * 8
    s = Template(RELATIONSHIP_CREATE_TEMPLATE)

    args = []

    name = c.obj.get_kwarg('name',None,(str),**rel)
    parent_string = c.obj.get_kwarg('parent_string',None,(str),**rel)
    if parent_string is None:
        parent_model = c.obj.get_kwarg(['parent_model','parent_table'],None,(str),**rel)
        parent_column = c.obj.get_kwarg(['parent_column'],None,(str),**rel)
        parent_string = f"{parent_model}.{parent_column}"
    on_delete = c.obj.get_kwarg('on_delete',None,(str),**rel)
    on_update = c.obj.get_kwarg('on_update',None,(str),**rel)
    child_column = c.obj.get_kwarg('child_column',None,(str),**rel)
    
    if isinstance(on_delete,(str)):
        args.append(f"on_delete='{on_delete}'")
    if isinstance(on_update,(str)):
        args.append(f"on_update='{on_update}'")

    arg_string_list = ""
    if len(args) > 0:
        arg_string_list = ', '.join(args)
        arg_string_list = f",{arg_string_list}"

    val = s.substitute(
        indent=indent,
        name=name,
        parent_string=parent_string,
        child_column=child_column,
        args=arg_string_list,
    )


    return val

UNIQUE_CONSTRAINT_TEMPLATE = """${indent}self.$name = UniqueConstraint(columns=[$columns])"""
def _gen_unique_constraint(rel:dict):
    indent = " " * 8
    s = Template(UNIQUE_CONSTRAINT_TEMPLATE)

    args = []

    name = c.obj.get_kwarg('name',None,(str),**rel)
    columns = c.obj.get_kwarg('columns',None,(list),**rel)
    # if parent_string is None:
    #     parent_model = c.obj.get_kwarg(['parent_model','parent_table'],None,(str),**rel)
    #     parent_column = c.obj.get_kwarg(['parent_column'],None,(str),**rel)
    #     parent_string = f"{parent_model}.{parent_column}"
    # on_delete = c.obj.get_kwarg('on_delete',None,(str),**rel)
    # on_update = c.obj.get_kwarg('on_update',None,(str),**rel)
    # child_column = c.obj.get_kwarg('child_column',None,(str),**rel)
    
    # if isinstance(on_delete,(str)):
    #     args.append(f"on_delete='{on_delete}'")
    # if isinstance(on_update,(str)):
    #     args.append(f"on_update='{on_update}'")

    column_list_string = ""
    if len(columns) > 0:
        cols = [f"self.{x}" for x in columns]
        column_list_string = ', '.join(cols)

    val = s.substitute(
        indent=indent,
        name=name,
        columns=column_list_string,
    )


    return val

FULLTEXT_TEMPLATE = """${indent}self.$name = FullTextIndex([$columns])"""
def _gen_fulltext_index(rel:dict):
    indent = " " * 8
    s = Template(FULLTEXT_TEMPLATE)

    # args = []

    name = c.obj.get_kwarg('name',None,(str),**rel)
    columns = c.obj.get_kwarg('columns',None,(list),**rel)

    column_list_string = ""
    if len(columns) > 0:
        cols = [f"self.{x}" for x in columns]
        column_list_string = ', '.join(cols)

    val = s.substitute(
        indent=indent,
        name=name,
        columns=column_list_string,
    )


    return val






td ={
    "name": "task_id",
    "data_type": {
        "sql_type_name": "INTEGER",
        "data_length": None,
        "python_type_name": "int",
        "class_name": "Integer"
    },
    "default": "no default",
    "unique": False,
    "fulltext": False,
    "nullable": False,
    "comment": None,
    "is_foreign_key": False,
    "is_primary": True,
    "auto_increment": True
}

td = {
        "name": "timestamp",
        "data_type": {
            "sql_type_name": "INTEGER",
            "data_length": None,
            "python_type_name": "int",
            "class_name": "Integer"
        },
        "default": None,
        "unique": False,
        "fulltext": False,
        "nullable": True,
        "comment": "The unix timestamp of when this was last modified, None otherwise.",
        "is_foreign_key": False,
        "is_primary": False,
        "auto_increment": False
    }

tdr =                         {
    "name": "parent_task",
    "fk_name": "FK_DQGjBpjafi0D",
    "on_delete": "CASCADE",
    "on_update": "CASCADE",
    "parent_model": "tasks",
    "parent_column": "task_id",
    "child_model": "tasks",
    "child_column": "parent_task_id"
}
tdu = {
    "name": "UQ_tasks_title",
    "comment": "Ensure the title is unique.",
    "columns": [
        "title"
    ]
}
tdft =                         {
                            "name": "FT_tasks_title",
                            "columns": [
                                "title"
                            ]
                        }


table =                 {
                    "name": "tasks",
                    "description": "Defines a Task Entity",
                    "columns": [
                        {
                            "name": "task_id",
                            "data_type": {
                                "sql_type_name": "INTEGER",
                                "data_length": None,
                                "python_type_name": "int",
                                "class_name": "Integer"
                            },
                            "default": "no default",
                            "unique": False,
                            "fulltext": False,
                            "nullable": False,
                            "comment": None,
                            "is_foreign_key": False,
                            "is_primary": True,
                            "auto_increment": True
                        },
                        {
                            "name": "title",
                            "data_type": {
                                "sql_type_name": "VARCHAR",
                                "data_length": 255,
                                "python_type_name": "str",
                                "class_name": "String"
                            },
                            "default": "no default",
                            "unique": True,
                            "fulltext": True,
                            "nullable": False,
                            "comment": None,
                            "is_foreign_key": False,
                            "is_primary": False,
                            "auto_increment": False
                        },
                        {
                            "name": "description",
                            "data_type": {
                                "sql_type_name": "VARCHAR",
                                "data_length": 500,
                                "python_type_name": "str",
                                "class_name": "String"
                            },
                            "default": "no default",
                            "unique": False,
                            "fulltext": False,
                            "nullable": False,
                            "comment": None,
                            "is_foreign_key": False,
                            "is_primary": False,
                            "auto_increment": False
                        },
                        {
                            "name": "start_timestamp",
                            "data_type": {
                                "sql_type_name": "INTEGER",
                                "data_length": None,
                                "python_type_name": "int",
                                "class_name": "Integer"
                            },
                            "default": None,
                            "unique": False,
                            "fulltext": False,
                            "nullable": True,
                            "comment": None,
                            "is_foreign_key": False,
                            "is_primary": False,
                            "auto_increment": False
                        },
                        {
                            "name": "end_timestamp",
                            "data_type": {
                                "sql_type_name": "INTEGER",
                                "data_length": None,
                                "python_type_name": "int",
                                "class_name": "Integer"
                            },
                            "default": None,
                            "unique": False,
                            "fulltext": False,
                            "nullable": True,
                            "comment": None,
                            "is_foreign_key": False,
                            "is_primary": False,
                            "auto_increment": False
                        },
                        {
                            "name": "timestamp",
                            "data_type": {
                                "sql_type_name": "INTEGER",
                                "data_length": None,
                                "python_type_name": "int",
                                "class_name": "Integer"
                            },
                            "default": None,
                            "unique": False,
                            "fulltext": False,
                            "nullable": True,
                            "comment": "The unix timestamp of when this was last modified, None otherwise.",
                            "is_foreign_key": False,
                            "is_primary": False,
                            "auto_increment": False
                        },
                        {
                            "name": "parent_task_id",
                            "data_type": {
                                "sql_type_name": "INTEGER",
                                "data_length": None,
                                "python_type_name": "int",
                                "class_name": "Integer"
                            },
                            "default": "no default",
                            "unique": False,
                            "fulltext": False,
                            "nullable": True,
                            "comment": None,
                            "is_foreign_key": True,
                            "is_primary": False,
                            "auto_increment": False
                        },
                        {
                            "name": "task_list_id",
                            "data_type": {
                                "sql_type_name": "INTEGER",
                                "data_length": None,
                                "python_type_name": "int",
                                "class_name": "Integer"
                            },
                            "default": "no default",
                            "unique": False,
                            "fulltext": False,
                            "nullable": False,
                            "comment": None,
                            "is_foreign_key": True,
                            "is_primary": False,
                            "auto_increment": False
                        },
                        {
                            "name": "modified_timestamp",
                            "data_type": {
                                "sql_type_name": "INTEGER",
                                "data_length": None,
                                "python_type_name": "int",
                                "class_name": "Integer"
                            },
                            "default": None,
                            "unique": False,
                            "fulltext": False,
                            "nullable": True,
                            "comment": "The unix timestamp of when this was last modified, None otherwise.",
                            "is_foreign_key": False,
                            "is_primary": False,
                            "auto_increment": False
                        },
                        {
                            "name": "deleted",
                            "data_type": {
                                "sql_type_name": "INTEGER",
                                "data_length": None,
                                "python_type_name": "int",
                                "class_name": "Integer"
                            },
                            "default": None,
                            "unique": False,
                            "fulltext": False,
                            "nullable": True,
                            "comment": "The unix timestamp of when this was deleted, None otherwise.",
                            "is_foreign_key": False,
                            "is_primary": False,
                            "auto_increment": False
                        }
                    ],
                    "relationships": [
                        {
                            "name": "parent_task",
                            "fk_name": "FK_DQGjBpjafi0D",
                            "on_delete": "CASCADE",
                            "on_update": "CASCADE",
                            "parent_model": "tasks",
                            "parent_column": "task_id",
                            "child_model": "tasks",
                            "child_column": "parent_task_id"
                        },
                        {
                            "name": "fk_tasklist_task",
                            "fk_name": "FK_bQR3YdseJ2J2",
                            "on_delete": "CASCADE",
                            "on_update": "CASCADE",
                            "parent_model": "task_lists",
                            "parent_column": "task_list_id",
                            "child_model": "tasks",
                            "child_column": "task_list_id"
                        }
                    ],
                    "unique_constraints": [
                        {
                            "name": "UQ_tasks_title",
                            "comment": "Ensure the title is unique.",
                            "columns": [
                                "title"
                            ]
                        }
                    ],
                    "fulltext_indexes": [
                        {
                            "name": "FT_tasks_title",
                            "columns": [
                                "title"
                            ]
                        }
                    ],
                    "parent_tables": [
                        "tasks",
                        "task_lists"
                    ],
                    "child_tables": [
                        "tasks"
                    ]
                }




if __name__ == '__main__':
    # result = _gen_column_create(td)
    # result = _gen_relationship(tdr)
    # result = _gen_unique_constraint(tdu)
    # result = _gen_fulltext_index(tdft)
    result = _gen_model_from_dict(table)
    
    
    print(f"result: {result}")


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable
import colemen_utils as c
import volent.settings as _settings
import volent.settings.types as _t




@dataclass
class TypeBase:
    data_length:int = None
    '''The column specific length for this data type'''
    
    sql_type_name:str = None
    '''The SQL name for the data type'''

    python_data_type:tuple = None
    '''The python equivalent data type'''
    
    max_data_length:int = None
    '''The maximum length of this data type'''
    min_data_length:int = None
    '''The minimum length of this data type'''

    def __init__(self,data_length:int=None) -> None:
        self.data_length = data_length

    @property
    def summary(self):
        '''
            Get this TypeBase's summary

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 08:34:13
            `@memberOf`: TypeBase
            `@property`: summary
        '''
        value = {
            "sql_type_name":self.sql_type_name,
            "data_length":self.data_length,
            "python_type_name":self.python_type_name,
            "class_name":self.__class__.__name__,
            # "max_data_length":self.max_data_length,
            # "min_data_length":self.min_data_length,
        }
        return value

    def _validate_data_length(self,data_length:int=None,default_value:int=None):
        if data_length is None:
            if default_value is None:
                self.data_length = data_length
                return
            data_length = default_value
        if isinstance(data_length,(float)):
            data_length = int(data_length)
        if isinstance(data_length,(int)) is False:
            raise TypeError(f"The data_length must be an integer")


        if self.max_data_length is not None and self.min_data_length is not None:
            if data_length > self.max_data_length or data_length < self.min_data_length:
                raise ValueError(f"The data_length for a {self.sql_type_name} must range from {self.min_data_length} to {self.max_data_length}")
        self.data_length = data_length

    @property
    def sql(self):
        '''
            Get this Type's sql data type

            BOOLEAN,VARCHAR(255)

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-24-2023 04:23:20
            `@memberOf`: Bool
            `@property`: mysql
        '''
        # if self.flavored_name in ["BIGINT","INTEGER"]:
        if self.data_length is None:
            return self.flavored_name
        return f"{self.flavored_name}({self.data_length})"

    @property
    def flavored_name(self):
        '''
            Get this TypeBase's flavored_name

            This is the sql_type_name converted to the sql flavor.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-24-2023 06:18:40
            `@memberOf`: TypeBase
            `@property`: flavored_name
        '''
        value = self.sql_type_name
        if _settings.globe.flavor == "sqlite":
            value = c.types.mysql_to_sqlite(self.sql_type_name)
        return value

    @property
    def python_type_name(self):
        '''
            Get this TypeBase's python_type_name

            This is the python type that is equivalent to this data type.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-24-2023 05:01:21
            `@memberOf`: TypeBase
            `@property`: python_type_name
        '''
        # value = self.python_data_type[0].__repr__
        # value = value.replace("<slot wrapper '__repr__' of '","")
        # value = value.replace("' objects>","")
        value = self.python_data_type.__name__
        # value = self.python_data_type[0]
        return value



    # def coerce(value):
    #     type_name = 





    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} : {self.python_type_name}>"
    
    
    
    



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


from dataclasses import dataclass
import re
from typing import Iterable,OrderedDict, Union


import colemen_utils as c


import volent.settings.types as _t
import volent.settings as _settings
from volent.Field import Field as _field
# from volent.Relationship import Relationship as _relationship
# from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint
from volent.query.Query import Query
from volent.query.WhereMixin import WhereMixin
from mysql.connector.errors import OperationalError,InterfaceError

@dataclass
class Select(Query,WhereMixin):
    def __init__(self,model:_t.model_type,columns=None,limit:int=100,offset:int=0) -> None:
        self._params = {}
        self._selects = []
        self._joins = []
        # self._limit = limit
        # self._offset = offset

        self.query_crud_type = "read"
        if columns is None:
            columns = model.columns
        if isinstance(columns,(tuple)):
            if len(columns) == 1:
                columns = columns[0]
        # print(f"Select.__init__.columns : {columns}")
        super().__init__(model,select_columns=columns)

    @property
    def has_joins(self):
        '''
            Get this Select's has_joins

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 05-07-2023 16:56:27
            `@memberOf`: Select
            `@property`: has_joins
        '''
        if len(self._joins) > 0:
            return True
        return False

    @property
    def select_string(self)->str:
        '''
            Get the compiled select string for this query.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-09-2022 15:56:15
            `@memberOf`: SelectQuery
            `@property`: select_string
        '''
        if len(self._selects) == 0:
            value = "*"
        else:
            selects = []
            for sel in self._selects:
                value = None
                if sel['alias'] is not None:
                    value = f"{sel['column_name']} as {sel['alias']}"
                else:
                    value = f"{sel['column_name']}"
                if value is not None:
                    selects.append(value)
            value = ', '.join(selects)

        if self._count is True:
            value = f"COUNT({value})"
        if self._average is True:
            value = f"AVG({value})"
        if self.sum_ is True:
            value = f"SUM({value})"
        return value


    @property
    def query(self):

        # print(f"self.select_string: {self.select_string}")

        value = f"""SELECT {self.select_string} FROM {self.model.quoted_name}{self.left_join_string}{self.where_string}"""

        # value = self._paginate_select_query(value)
        # value = self._format_query_params(value,self._params)
        # print(f"value: {value}")
        # print(f"self._params: {self._params}")
        return value,self.serialized_params

    def execute(self,return_models=True)->Union[bool,dict,int,_t.model_type]:
        '''
            Execute the select query on the table.

            ----------

            Arguments
            -------------------------
            [`return_models`=True] {bool}
                If False, this will return a dictionry or list of dictionaries.


            Return {bool,dict,list,int,model}
            ----------------------
            returns False if the query is malformed or could not be executed.
            The results of the query otherwise.

            If return_models is True, it will generate a new model instance for each row.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:54:49
            `memberOf`: Select
            `version`: 1.0
            `method_name`: execute
            * @xxx [04-28-2023 07:57:23]: documentation for execute
        '''
    # def execute(self)->Union[bool,dict,int]:
        # _mysql_connector.MySQLInterfaceError: Lost connection to MySQL server during query
        sql,args= self.query

        if sql is False:
            return False

        # print(f"sql:{sql}")
        # print(f"args:{args}")
        try:
            # @Mstep [] execute the insert query.

            # result = self.database.run_select(sql,args)
            result = self.database.run_select_internal(sql,self,args)

            # print(f"result : {type(result)} : {result}")
            if return_models and result is not None:
                models = []
                # if isinstance(result[0],(tuple)) is False:
                #     result = [result]
                for r in result:
                    # print(f"r : {r}")
                    mdl = self.model.__class__(**r)
                    mdl.__doc__ = self.model.__doc__
                    mdl._model_name = self.model._model_name
                    mdl._database_name = self.model._database_name
                    mdl._database = self.model._database
                    models.append(mdl)
                return models
        except OperationalError as e:
            c.con.log(e,"red")
        except InterfaceError as e:
            c.con.log(e,"red")
        return result

    def count(self)->_t.query_type:
        '''
            Count the number of rows that are selected.

            ----------


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:57:50
            `memberOf`: Select
            `version`: 1.0
            `method_name`: count
            * @xxx [04-28-2023 07:58:26]: documentation for count
        '''
        self._count = True
        return self

    def average(self)->_t.query_type:
        '''
            Calculate the average of all the rows that are selected.

            ----------


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:58:34
            `memberOf`: Select
            `version`: 1.0
            `method_name`: average
            * @xxx [04-28-2023 07:59:02]: documentation for average
        '''
        self._average = True
        return self

    def sum_(self)->_t.query_type:
        '''
            Calculate the sum of all the rows that are selected.

            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:59:07
            `memberOf`: Select
            `version`: 1.0
            `method_name`: sum_
            * @xxx [04-28-2023 07:59:29]: documentation for sum_
        '''
        self._sum = True
        return self

    def add_select(self,column_name,alias=None)->_t.query_type:
        '''
            Add a column to the selection of this query.

            By default "*" is used to select all columns if None are specified.
            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to select
            [`alias`=None] {str}
                An alias to use for this column

            Return {None}
            ----------------------
            returns nothing

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-13-2022 12:04:31
            `memberOf`: SelectQuery
            `version`: 1.0
            `method_name`: add_select
            * @xxx [12-13-2022 12:06:00]: documentation for add_select
        '''
        if self.model is not None:
            if self.model.get_column(column_name) is None:
                c.con.log(f"Column {column_name} does not exist in table: {self.model.name}","warning")
                return
        data = {
            "column_name":column_name,
            "alias":alias,
        }
        self._selects.append(data)
        return self

    def left_join(self,model:_t.model_type,column_name:str):
        data = {
            "join_type":"LEFT",
            "model":model,
            "table_name":model.quoted_name,
            "column_name":column_name,
        }
        self._joins.append(data)
    # def left_join(self,table_name:str,column_name:str):
    #     data = {
    #         "join_type":"LEFT",
    #         "table_name":table_name,
    #         "column_name":column_name,
    #     }
    #     self._joins.append(data)


    @property
    def left_join_string(self):
        '''
            Get this Select's left_join_string

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-21-2023 08:58:54
            `@memberOf`: Select
            `@property`: left_join_string
        '''
        out_list = []
        for join in self._joins:
            if join['join_type'] != "LEFT":
                continue
            template = f""" {join['join_type']} JOIN {self.model.quoted_name}
    ON {self.model.quoted_name}.{join['column_name']} = {join['table_name']}.{join['column_name']}"""
            return template
        return ""


    def paginate(self,page_number:int=1,per_page:int=None):
        '''
            update the pagination settings for this query.

            Arguments
            -------------------------
            `page_number` {int} = 1
                Which page of items to return

                If set to 0, None or False the query will NOT be paginated.

            `per_page` {int} = None
                How many items per page should be shown.

            Return {Select}
            ----------------------
            returns this select instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-02-2024 10:14:25
            `memberOf`: Select
            `version`: 1.0
            `method_name`: paginate
            * @xxx [04-02-2024 10:16:07]: documentation for paginate
        '''
        self._page_number = page_number
        if per_page is not None:
            self._items_per_page = per_page
        return self

    def _paginate_query(self,sql:str)->str:
        page_number = self._page_number
        per_page = self._items_per_page
        offset = None if (page_number * per_page) < 1 else page_number * per_page
        limit = per_page

        if limit in [0,None,False]:
            return sql

        # @Mstep [] remove any trailing spaces and semicolons
        sql = c.string.strip(sql,[";"," "],"right")
        # @Mstep [] remove any already existing pagination strings.
        sql = re.sub(r'limit\s*[0-9]*\s*(,|offset)\s*(:?[0-9\s]*)?',"",sql,re.MULTILINE | re.IGNORECASE)

        limit_string = f"LIMIT {limit}"
        offset_string = ""
        if offset is not None:
            offset_string = f"OFFSET {offset}"

        paginate = f"{limit_string} {offset_string}"
        sql = f"{sql} {paginate}"
        return sql


    def __repr__(self) -> str:
        return f"<Select : {self.model.model_name}>"


def format_null(value):
    if value is None:
        return "NULL"
    return value




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable
import colemen_utils as c
import volent.settings as _settings
import volent.settings.types as _t
from volent.data_types.TypeBase import TypeBase as _type_base
from volent.exceptions import ValidationError


@dataclass
class String(_type_base):


    def __init__(self,data_length:int=None) -> None:
        super().__init__(data_length)
        self.min_data_length = 0
        self.max_data_length = 65535
        self.sql_type_name = "VARCHAR"
        self.python_data_type = (str)

        self._validate_data_length(data_length,_settings.control.varchar_default_length)

    @property
    def mysql(self):
        '''
            Get this String's mysql

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-24-2023 04:23:20
            `@memberOf`: String
            `@property`: mysql
        '''
        value = f"varchar({self.data_length})"
        return value

    def _serialize(self,value):
        if isinstance(value,(str)):
            return value

        if isinstance(value,(int,float)):
            return str(value)

        if isinstance(value,(bool)):
            return c.types.bool_to_string(value)

    def _deserialize(self,value):
        return value

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} : {self.data_length}>"


    def serialized_value(self,value,field_name:str=None)->str:

        if isinstance(value,(str)):
            return value

        if isinstance(value,(int,float)):
            return str(value)

        if isinstance(value,(bool)):
            return c.types.bool_to_string(value)

        if isinstance(value,(str)) is False:
            raise ValidationError(f"Failed to serialize {value} to a string.",field_name)

        return value

    def deserialized_value(self,value,field_name:str=None):

        if value == _t.undefined or value == _t.no_default:
            return value
        if value is None:
            return value
        
        if isinstance(value,(str)):
            result = c.types.to_bool(value,default=None)
            if result is not None:
                return result

            result = c.types.to_number(value)
            if result is not None:
                return result
            
            return value


        if isinstance(value,(bool)) is False:
            raise ValidationError(f"Failed to deserialize {value}.",field_name)

        return value



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable, Union
import colemen_utils as c
import volent.settings as _settings
import volent.settings.types as _t
from volent.mixins.MySQLGeneratorMixin import MySQLGeneratorMixin

@dataclass
class UniqueConstraint(MySQLGeneratorMixin):
    main:_t._main_type = None
    # database:_t.database_type = None
    model:_t.model_type = None
    name:str = None
    comment:str = None
    columns:Iterable[_t.column_type] = None
    comment:str = None




    def __init__(
        self,
        columns:Iterable[_t.column_type],
        name:str=None,
        comment:str=None,
        ):
        '''
            Create a unique constraint instance.

            ----------

            Arguments
            -------------------------
            `Columns` {list,column}
                A column or list of columns to create a unique constraint on.

            `name` {str}
                The name of this unique constraint. This name must be unique through out the database.

            `comment` {str}
                The comment to add to the unique constraint in the database.


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 09:59:09
            `memberOf`: UniqueConstraint
            `version`: 1.0
            `method_name`: UniqueConstraint
            * @xxx [04-28-2023 10:01:29]: documentation for UniqueConstraint
        '''
        self.name = name
        self.comment = comment
        self.columns = c.arr.force_list(columns)



        # _parse_parent(self,parent)


    @property
    def summary(self):
        '''
            Get this UniqueConstraint's summary

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-23-2023 14:53:39
            `@memberOf`: UniqueConstraint
            `@property`: summary
        '''
        value = {
            "name":self.name,
            "comment":self.comment,
            "columns":[x.name for x in self.columns],
        }
        return value



    def __hash__(self):
        col_names = [x.name for x in self.columns]
        val = col_names.append(self.name)
        return hash(val)

    def __eq__(self,other):
        if isinstance(other,UniqueConstraint):
            return self.__hash__() == other.__hash__()
        return False




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


from typing import Iterable, OrderedDict, Union
from collections import OrderedDict
import colemen_utils as c


class OrderedClass(type):
    @classmethod
    def __prepare__(mcs, name, bases): 
        return OrderedDict()




    def __setattr__(self, name, value):
        super().__setattr__(name, value)

        if not name in self._order:
            self._order.append(name)



        return value

    def ordered_attrs(self, with_order=False):
        return [(k,getattr(self, k)) for k in self._order if k != '_order' or with_order]


    def __new__(cls, name, bases, classdict):
        result = type.__new__(cls, name, bases, dict(classdict))
        result._order = []
        return result

    # def __new__(cls, name, bases, classdict):
    #     # print(f"classdict:{classdict}")
    #     result = type.__new__(cls, name, bases, dict(classdict))
    #     result.__odict__ = OrderedDict()
    #     result.__odict__ = list(classdict.keys())
    # #     print(f"result.__init__.__dict__:{result.__init__.}")
    # #     # result.__columns__ = []
    # #     print("-"*80)
    # #     print(result.__dict__)
    # #     print("-"*80)

    # #     # value = []
    # #     # from volent.Column import Column as _column
    # #     # for prop in list(classdict.keys()):
    # #     #     name = prop

    # #     #     val = classdict[prop]
    # #     #     # print(f"{prop} - prop type: {type(val)}")
    # #     #     if isinstance(val,(_column)):
    # #     #         c.con.log(f"OrderedClass.located Column: {name}","green")
    # #     #         val.name = name
    # #     #         val.model = result
    # #     #         if val.is_primary is True:
    # #     #             result.primary_column = val
    # #     #         value.append(val)
                
    # #     #         setattr(result,name,val)

            
    # #         # if isinstance(prop,(list)):
    # #         #     # print(f"prop is a list")
    # #         #     vals = classdict[prop]
    # #         #     for val in vals:
    # #         #         print(f"val:{val}")
    # #         #         if val.name == name:
    # #         #     # val = classdict[prop]
    # #         #     # print(f"{name} - val: {value}")
    # #         #             if isinstance(val,(_column)):
    # #         #                 c.con.log(f"OrderedClass.located Column: {name}","green")
    # #         #                 val.name = name
    # #         #                 val.model = result
    # #         #                 if val.is_primary is True:
    # #         #                     result.primary_column = val
    # #         #                 value.append(val)
    # #         #                 setattr(result,name,val)

        
    # #     # if len(value) == 0:
    # #     #     value = None
        
    # #     # result.title.value = "boobers"
    # #     # result._columns = value
    # #     # result.boobs = "tiddies"
    # #     return result
    
    
    
    # # def __new__(cls, *args, **kwargs):
    # #     instance = type.__new__(cls)
    # #     instance.__odict__ = OrderedDict()
    # #     return instance

    # def __setattr__(self, key, value):
    #     if key != '__odict__':
    #         self.__odict__[key] = value
    #     object.__setattr__(self, key, value)

    # def keys(self):
    #     return self.__odict__.keys()

    # def iteritems(self):
    #     return self.__odict__.iteritems()





# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
import re
from typing import Iterable, OrderedDict, Sized, Union
from collections import OrderedDict
import volent.settings.types as _t
import colemen_utils as c

# @dataclass
# class Trigger:
#     model:_t.model_type = None
#     when:str = None

#     def __init__(
#         self,
#         model:_t.model_type,
#         when:str,
#         ):
#         self.model = model
#         self.when = when




#     @property
#     def create_statement(self):
#         '''
#             Get this Trigger's create_statement

#             `default`:None


#             Meta
#             ----------
#             `@author`: Colemen Atwood
#             `@created`: 05-08-2023 11:28:54
#             `@memberOf`: Trigger
#             `@property`: create_statement
#         '''
#         value = self.create_statement
#         if value is None:
#             value = somethingGoesHere
#             self.create_statement = value
#         return value
    




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
'''
    The relationship module.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 04-28-2023 10:06:56
    `name`: relationship
    * @xxx [04-28-2023 10:07:06]: documentation for relationship
'''


from dataclasses import dataclass
from typing import Iterable, Union
import colemen_utils as c
import volent.settings as _settings
import volent.settings.types as _t

from volent.mixins.MySQLGeneratorMixin import MySQLGeneratorMixin

@dataclass
class Relationship(MySQLGeneratorMixin):
    main:_t._main_type = None
    # database:_t.database_type = None
    model:_t.model_type = None
    name:str = None
    '''The name of this relationship'''
    on_delete:str = None
    '''The action to perform with the parent is deleted'''
    on_update:str = None
    '''The action to perform with the parent is updated'''
    fk_name:str = None
    '''The name used to uniquely identify this relationship in the database.'''

    # _parent_database_name:str = None
    # _parent_model_name:str = None
    # _parent_column_name:str = None

    parent_model:_t.model_type = None
    '''A reference to the parent model instance'''
    parent_column:_t.column_type = None
    '''A reference to the parent model's column instance'''

    child_model:_t.model_type = None
    '''A reference to the child model instance'''
    child_column:_t.model_type = None
    '''A reference to the child model's column instance'''
    # schemas:Iterable[_t.schema_type] = None



    def __init__(
        self,
        child:Union[str,_t.column_type],
        parent:str,
        name:str=None,
        on_delete:str = None,
        on_update:str = None,
        fk_name:str = None,
        ):
        '''
            Create a relationship instance.

            ----------

            Arguments
            -------------------------
            `child` {column}
                The child column of this relationship

            `parent` {str}
                The name (dot path) of the parent column that this relationship represents.

                "tasks.task_id"

            [`name`=None] {str}
                The name of this relationship, if not provided, the variable name is used from its declaration
                in the model.

            [`on_delete`=None] {str}
                The action to perform when the parent row is deleted.

            [`on_update`=None] {str}
                The action to perform when the parent row is updated.

            [`fk_name`=None] {str}
                The name used to uniquely identify this relationship in the database.
                If not provided, a random one will be generated.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 10:07:32
            `memberOf`: Relationship
            `version`: 1.0
            `method_name`: relationship
            * @xxx [04-28-2023 10:11:24]: documentation for relationship
        '''
        self.name = name
        self.on_delete = on_delete
        self.on_update = on_update
        self.parent = parent
        self.child_column = child
        self.fk_name = f"FK_{c.rand.rand()}" if fk_name is None else fk_name




    @property
    def summary(self):
        '''
            Get this Relationship's summary

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-23-2023 14:53:39
            `@memberOf`: Relationship
            `@property`: summary
        '''
        value = {
            "name":self.name,
            "fk_name":self.fk_name,
            "on_delete":self.on_delete,
            "on_update":self.on_update,
            "parent_model":self.parent_model.name,
            "parent_column":self.parent_column.name,
            "child_model":self.child_model.name,
            "child_column":self.child_column.name,
            "parent_string":self.parent,
        }
        return value


    # def locate_parent(self):
    #     database = self._parent_database_name
    #     model = self._parent_model_name
    #     column = self._parent_column_name
    #     if database is not None:
    #         self.main






# def _parse_parent(relationship:Relationship,value):
#     value = c.string.strip_excessive_chars(value,["."])
#     pl = value.split(".")
#     if len(pl) == 3:
#         relationship._parent_database_name = pl[0]
#         relationship._parent_model_name = pl[1]
#         relationship._parent_column_name = pl[2]
#     if len(pl) == 2:
#         relationship._parent_model_name = pl[0]
#         relationship._parent_column_name = pl[1]
#     if len(pl) == 1:
#         relationship._parent_column_name = pl[0]


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


from dataclasses import dataclass
import re
from typing import Iterable,OrderedDict, Union


import colemen_utils as c


import volent.settings.types as _t
import volent.settings as _settings
from volent.Field import Field as _field
# from volent.Relationship import Relationship as _relationship
# from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint
from volent.query.Query import Query
from volent.query.WhereMixin import WhereMixin


@dataclass
class Update(Query,WhereMixin):
    def __init__(self,model:_t.model_type,columns=None) -> None:
        self._params = {}
        self._updates = []

        self.data = columns
        self.ignore_on_updates = False

        self.limit = 1
        '''Limit the number of rows that can be updated by this query.

        if set to 0, None or False, no limit will be applied.
        '''

        self.query_crud_type = "update"
        super().__init__(model,update_columns=columns)

    @property
    def select_string(self)->str:
        '''
            Get the compiled select string for this query.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-09-2022 15:56:15
            `@memberOf`: SelectQuery
            `@property`: select_string
        '''
        if len(self._selects) == 0:
            value = "*"
        else:
            selects = []
            for sel in self._selects:
                value = None
                if sel['alias'] is not None:
                    value = f"{sel['column_name']} as {sel['alias']}"
                else:
                    value = f"{sel['column_name']}"
                if value is not None:
                    selects.append(value)
            value = ', '.join(selects)

        if self._count is True:
            value = f"COUNT({value})"
        if self._average is True:
            value = f"AVG({value})"
        if self.sum_ is True:
            value = f"SUM({value})"
        return value

    @property
    def query(self):
        '''
            Generate the query string for this update.

            Return {tuple}
            ----------------------
            A tuple containing the query string and a dictionary of parameters

            If no parameters are used, its just an empty dict.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-02-2024 09:50:49
            `memberOf`: Update
            `version`: 1.0
            `method_name`: query
            * @xxx [04-02-2024 09:51:49]: documentation for query
        '''
        # print(self.data)
        self.updates_from_data()

        value = f"UPDATE {self.model.quoted_name} SET {self.update_string}{self.where_string}"

        value = self._paginate_query(value)
        # value = self._format_query_params(value,self._params)
        return value,self.serialized_params
        # return value,self._params

    def execute(
        self,
        foreign_key_checks:bool=True
        )->Union[bool,dict,int]:
        '''
            Execute this update operation on the model.

            ----------

            Arguments
            -------------------------

            [`foreign_key_checks`=True] {bool}
                If False, the query will not perform foreign key checks before executing the update.
                Be careful, this can cause integrity issues and will NOT warn you about it.


            Return {type}
            ----------------------
            return_description

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 08:01:22
            `memberOf`: Update
            `version`: 1.0
            `method_name`: execute
            * @TODO []: documentation for execute
        '''
        sql,args= self.query

        if sql is False:
            return False

        # print(f"sql:{sql}")
        # print(f"args:{args}")

        # @Mstep [] execute the update query.
        result = self.database.run(sql,args,foreign_key_checks=foreign_key_checks)
        # @Mstep [] select the updated row as a dictionary.
        data = self.model.select().copy_wheres(self).execute(return_models=False)
        data = data[0]
        # @Mstep [LOOP] iterate the dictionary.
        for k,v in data.items():
            # @Mstep [] retrieve the column with a matching name from the model.
            col = self.model.get_column(k)
            col.value = v
            self.model._saved = True
        return data
        models = []
        for r in result:
            mdl = self.model.__class__(**r)
            mdl.__doc__ = self.model.__doc__
            mdl._name = self.model.name
            mdl._database_name = self.model._database_name
            mdl._database = self.model._database
            models.append(mdl)
        return models
        return result

    def updates_from_data(self)->dict:
        '''
        Parse this query's data to prepare it for updating the database.

        This will:
        - filter by existing columns
        - remove undefined values
        - apply on_update values

        returns a dictionary and sets the self._updates value.
        '''
        if self.data is None:
            self.data = self.model.data

        data = self.filter_dict_by_columns(self.data)
        updates = self.filter_undefined_values(data)

        if self.ignore_on_updates is False:
            updates = self.set_on_update_values(updates)
        self._updates = updates

        return updates



    @property
    def update_string(self):
        '''
            generate a string of placeholder values that will update columns.

            `example`: track_number=@track_number, genre=@genre

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-09-2022 15:56:15
            `@memberOf`: UpdateQuery
            `@property`: update_string
        '''
        if len(self._updates) == 0:
            return None
        else:
            placeholders = []
            for k,v in self._updates.items():
                p = self.add_param(k,v)
                placeholders.append(p.parameterized_update_string)
            value = ', '.join(placeholders)
        return value


    def add_update(self,column_name:str,value)->_t.update_query_type:
        '''
            add a column and its value to be updated with.

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to update

            `value` {any}
                The value to update the column with.

            Return {Update}
            ----------------------
            returns this query instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-02-2024 10:27:40
            `memberOf`: Update
            `version`: 1.0
            `method_name`: add_update
            * @xxx [04-02-2024 10:28:42]: documentation for add_update
        '''
        if self.data is None:
            self.data = {}
        self.data[column_name] = value
        return self
        # self._updates[column_name] = value

    def __repr__(self) -> str:
        return f"<Update : {self.model.model_name}>"


    def paginate(self,limit:Union[int,bool,None]=1)->_t.update_query_type:
        '''
            Set the pagination limit for this query.


            `Remember`: by default SQLITE does not support limiting updates/deletes so this will have no effect.

            Arguments
            -------------------------
            `limit` {int,bool,None}
                The number of rows that this query is allowed to update.

                If the limit is set to 0, None or False, no pagination will be added.

            Return {Update}
            ----------------------
            returns this query instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-02-2024 10:25:38
            `memberOf`: Update
            `version`: 1.0
            `method_name`: paginate
            * @xxx [04-02-2024 10:26:56]: documentation for paginate
        '''
        self.limit = limit
        return self


    def _paginate_query(self,sql:str)->str:
        '''
            Apply the pagination values to the query string.

            Because this is for an update query only the limit value is used.

            If the limit is set to 0, None or False, no pagination will be added.

            `Remember`: by default SQLITE does not support limiting updates/deletes so this will have no effect.

            Arguments
            -------------------------
            `sql` {string}
                The query string to apply pagination to.

            Return {str}
            ----------------------
            The sql with pagination added.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-02-2024 10:20:39
            `memberOf`: Update
            `version`: 1.0
            `method_name`: _paginate_query
            * @xxx [04-02-2024 10:25:28]: documentation for _paginate_query
        '''
        if _settings.globe.flavor in ["sqlite"]:
            return sql

        limit = self.limit
        offset = 0


        # @Mstep [] remove any trailing spaces and semicolons
        sql = c.string.strip(sql,[";"," "],"right")
        # @Mstep [] remove any already existing pagination strings.
        sql = re.sub(r'limit\s*[0-9]*\s*(,|offset)\s*(:?[0-9\s]*)?',"",sql,re.MULTILINE | re.IGNORECASE)

        if limit in [0,None,False]:
            return sql

        limit_string = f"LIMIT {limit}"
        offset_string = f"OFFSET {offset}"

        paginate = f"{limit_string} {offset_string}"
        paginate = f"{limit_string}"
        sql = f"{sql}\n {paginate}"
        return sql





def format_null(value):
    if value is None:
        return "NULL"
    return value




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


from dataclasses import dataclass
from time import time
from typing import Iterable

import colemen_utils as _c

import volent.settings.types as _t
import volent.settings as _settings
from volent.mixins import MySQLGeneratorMixin
from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint
from volent.FullTextIndex import FullTextIndex as _fullTextIndex

@dataclass
class Column(MySQLGeneratorMixin):
    model:_t.model_type = None

    name:str = None
    data_type:_t.type_base_type = None
    default = _settings.types.no_default
    on_update = _t.undefined
    '''The value to set this column to when an update is performed'''
    on_insert = _t.undefined
    '''The value to set this column to when an insertion is performed'''
    nullable:bool = None
    comment:str = None
    is_foreign_key:bool = None
    is_primary:bool = None
    auto_increment = None
    _unique:bool = None
    _fulltext:bool = None
    is_private:bool = None
    dump_only:bool = None

    # relationship:Iterable[_t.relationship_type] = None
    # _column_value = "__NO_VALUE__"
    _column_value = _t.undefined

    def __init__(
        self,
        data_type:str,
        name:str=None,
        nullable:bool=False,
        comment:str=None,
        is_foreign_key:bool=False,
        is_primary:bool=False,
        auto_increment:bool=False,
        unique:bool=False,
        fulltext:bool=False,
        default=_settings.types.no_default,
        dump_only=False,
        on_update=_t.undefined,
        on_insert=_t.undefined,
        ):
        '''
            Create a column instance.

            ----------

            Arguments
            -------------------------
            `data_type` {str,data_type}
                The data_type that this column uses.
            [`name`=None] {str}
                The name of this column, if not provided, the variable name will be used from its declaration in the model

            [`nullable`=False] {bool}
                If True, the column will allow "NULL" values

            [`comment`=None] {str}
                The comment to add to the column in the database.

            [`is_foreign_key`=False] {bool}
                If True, this column should have an associated Relationship

            [`is_primary`=False] {bool}
                If True, this column will be treated as the primary index of the table.
                There can only be one of these.

            [`auto_increment`=False] {bool}
                if True, the column will be auto_incremented in the database.
                There can only be one of these

            [`unique`=False] {bool}
                If True, a unique constraint will be added to this column.

            [`fulltext`=False] {bool}
                If True, a fulltext index will be added to this column.

            [`default`=noDefault] {any}
                The default value to assign to this column if a value is not provided.

            [`dump_only`=False] {bool}
                If True, this column cannot be updated after insertion

            [`on_update`=undefined] {any}
                This value will be automatically set any time this column's row is updated.

            [`on_insert`=undefined] {any}
                This value will be automatically set when the column is inserted


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 10:12:26
            `memberOf`: Column
            `version`: 1.0
            `method_name`: column
            * @xxx [04-28-2023 10:19:39]: documentation for column
        '''

        self.name = name
        self.data_type = data_type
        self.default = default
        self.on_update = on_update
        self.on_insert = on_insert
        self._unique = unique
        self._fulltext = fulltext
        self.nullable = nullable

        self.comment = comment
        self.is_foreign_key = is_foreign_key
        self.is_primary = is_primary
        self.auto_increment = auto_increment
        self.dump_only = dump_only
        self._column_value = _t.undefined



    @property
    def summary(self):
        '''
            Get this Column's summary

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 08:26:42
            `@memberOf`: Column
            `@property`: summary
        '''
        value = {
            "name":self.name,
            "data_type":self.data_type.summary,
            "default":self.default,
            "unique":self.unique,
            "fulltext":self.fulltext,
            "nullable":self.nullable,
            "comment":self.comment,
            "is_foreign_key":self.is_foreign_key,
            "is_primary":self.is_primary,
            "auto_increment":self.auto_increment,
            "dump_only":self.dump_only,
            "on_update":self.on_update,
            "on_insert":self.on_insert,
        }
        if self.default == _t.no_default:
            value['default'] = "no default"
        if self.on_update == _t.undefined:
            value['on_update'] = "no_on_update"
        if self.on_insert == _t.undefined:
            value['on_insert'] = "no_on_insert"
        return value

    def __call__(self, *args, **kwds):
        return self._column_value


    @property
    def dot_path(self):
        '''
            Get this Column's dot_path

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-14-2023 07:22:26
            `@memberOf`: Column
            `@property`: dot_path
        '''
        value = self.name
        if self.model is not None:
            value = f"{self.model.name}.{self.name}"
        return value


    @property
    def value(self):
        '''
            Get this Column's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 14:49:51
            `@memberOf`: Column
            `@property`: value
        '''
        value = self._column_value
        if value == _t.undefined:
            if self.default != _t.no_default:
                value = self.default
                self._column_value = value
        return value

    @value.setter
    def value(self,value):
        '''
            Set the Column's value property

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-11-2023 14:54:15
            `@memberOf`: Column
            `@property`: value
        '''
        if value != self._column_value:
            self.model._saved = False
        self._column_value = value

    @property
    def serialized_value(self):
        '''
            Get this Column's serialized_value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-27-2023 11:05:37
            `@memberOf`: Column
            `@property`: serialized_value
        '''
        value = self.value
        if hasattr(self.data_type,"__serialize"):
            return self.data_type.__serialize(value)
        return value

    @property
    def deserialized_value(self):
        '''
            Get this Column's deserialized_value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 03-27-2023 11:05:37
            `@memberOf`: Column
            `@property`: serialized_value
        '''
        value = self.value
        if hasattr(self.data_type,"serializer"):
            return self.data_type.serializer(value,self.name)
        if hasattr(self.data_type,"__deserialize"):
            return self.data_type.__deserialize(value)
        if hasattr(self.data_type,"deserialized_value"):
            return self.data_type.deserialized_value(value)
        return value

    @property
    def unique(self):
        '''
            Get this Column's unique value.
            
            This will also create a unique constraint instance

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-28-2023 10:23:37
            `@memberOf`: Column
            `@property`: unique
        '''
        value = self._unique
        if value is True:
            # name = f"UQ_{_c.rand.rand()}"
            name = f"UQ_{self.model.model_name}_{self.name}"
            # if self.model.get_unique_constraint(name) is not None:
            #     return value
            comment = f"Ensure the {self.name} is unique."
            u = _uniqueConstraint(self,name,comment)
            # self.model.__setattr__(name,u)
            setattr(self.model,name,u)
        return value

    @property
    def fulltext(self):
        '''
            Get this Column's fulltext value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-28-2023 10:23:37
            `@memberOf`: Column
            `@property`: unique
        '''
        value = self._unique
        if value is True:
            name = f"FT_{self.model.model_name}_{self.name}"
            u = _fullTextIndex(self,name)
            setattr(self.model,name,u)
        return value

    def __repr__(self):
        return f"<{self.__class__.__name__} : {self.name} {self.data_type.python_data_type} {self.value}>"



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


from dataclasses import dataclass
from typing import Iterable, Union


import colemen_utils as c


import volent.settings.types as _t
from volent.query.WhereClause import WhereClause as _w
# import volent.settings as _settings
# from volent.Field import Field as _field
# from volent.Relationship import Relationship as _relationship
# from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint
# from volent.query.Query import Query


@dataclass
class WhereMixin:

    def equals(self,column_name:str,value,ignore_nulls:bool=False)->_t.query_type:
        '''
            Add an equals where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            `value` {any}
                The value to test for.

            [`ignore_nulls`=False] {bool}
                If True, the clause will not be added if the value is None.

            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: equals
            * @xxx [04-25-2023 11:18:26]: documentation for equals
        '''
        if value is None and ignore_nulls is True:
            return self
        if value == _t.no_default:
            return self
        value = format_null(value)

        self.add_where(column_name=column_name,value=value,comparison="is")
        return self
    is_ = equals

    def is_not(self,column_name:str,value)->_t.query_type:
        '''
            Add an "IS NOT" where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            `value` {any}
                The value to test for.

            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: equals
            * @xxx [04-25-2023 11:18:26]: documentation for equals
        '''
        value = format_null(value)
        self.add_where(column_name=column_name,value=value,comparison="is_not")
        return self
    not_ = is_not

    def null(self,column_name:str)->_t.query_type:
        '''
            Add a "IS NULL" where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: equals
            * @xxx [04-25-2023 11:18:26]: documentation for equals
        '''
        self.add_where(column_name=column_name,value="NULL",comparison="IS")
        return self

    def not_null(self,column_name:str)->_t.query_type:
        '''
            Add an "IS NOT NULL" where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: equals
            * @xxx [04-25-2023 11:18:26]: documentation for equals
        '''
        self.add_where(column_name=column_name,value="NULL",comparison="IS NOT")
        return self

    def in_(self,column_name:str,options:Iterable[Union[str,int,float]])->_t.query_type:
        '''
            Add an "IN" where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            `options` {list}
                The value to test for.

            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: in_
            * @xxx [04-25-2023 11:18:26]: documentation for in_
        '''
        self.add_where(column_name,options,"in")
        return self

    def not_in(self,column_name:str,options:list)->_t.query_type:
        '''
            Add a "NOT IN" where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            `options` {any}
                A list of options to compare against.

            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: not_in
            * @xxx [04-25-2023 11:18:26]: documentation for not_in
        '''
        # options = f"({','.join(c.arr.values_to_strings(options))})"
        self.add_where(column_name,options,"not in")
        return self

    def between(self,column_name:str,minimum:Union[int,float],maximum:Union[int,float])->_t.query_type:
        '''
            Add a "BETWEEN" where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            `minimum` {int,float}
                The minimum value
            `maximum` {int,float}
                The maximum value

            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: between
            * @xxx [04-25-2023 11:18:26]: documentation for between
        '''
        self.add_where(column_name,(minimum,maximum),"between")
        return self

    def greater_than(
        self,
        column_name:str,
        value:Union[int,float],
        inclusive:bool=False
        )->_t.query_type:
        '''
            Add a ">" where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            `value` {any}
                The value to test for.

            [`inclusive`=False] {bool}
                If True, the value can be less than or equal to, this is the same as calling
                greater_than_equal()



            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: greater_than
            * @xxx [04-25-2023 11:18:26]: documentation for greater_than
        '''
        if inclusive is True:
            self.greater_than_equal(column_name,value)
        else:
            self.add_where(column_name,value,"greater_than")
        return self

    def greater_than_equal(self,column_name:str,value:Union[int,float])->_t.query_type:
        '''
            Add a ">=" where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            `value` {any}
                The value to test for.

            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: greater_than_equal
            * @xxx [04-25-2023 11:18:26]: documentation for greater_than_equal
        '''
        self.add_where(column_name,value,"greater_than_equal")
        return self

    def less_than(self,column_name:str,value:Union[int,float],inclusive:bool=False)->_t.query_type:
        '''
            Add a "<" where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            `value` {int,float}
                The value that the column must be less than.

            [`inclusive`=False] {bool}
                If True, the value can be less than or equal to, this is the same as calling
                less_than_equal()

            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: less_than
            * @xxx [04-25-2023 11:18:26]: documentation for less_than
        '''
        if inclusive is True:
            self.less_than_equal(column_name,value)
        else:
            self.add_where(column_name,value,"less_than")
        return self

    def less_than_equal(self,column_name:str,value:Union[int,float])->_t.query_type:
        '''
            Add a "<=" where clause to the query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to test.

            `value` {int,float}
                The value that the column must be less than or equal to.

            Return {Query}
            ----------------------
            returns this Query Instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 11:16:43
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: less_than
            * @xxx [04-25-2023 11:18:26]: documentation for less_than
        '''
        self.add_where(column_name,value,"less_than_equal")
        return self

    def timestamps(self,column_name:str,start_timestamp:int=None,end_timestamp:int=None):
        '''
            Add a where clause used for filtering rows by dates.

            If both timestamps are provided only rows that are between will be returned.

            If only start is provided, the row timestamp must be greater.
            If only end is provided, the row timestamp must be less.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column to filter

            [`start_timestamp`=None] {int}
                The start_timestamp to filter by

            [`end_timestamp`=None] {int}
                The end_timestamp to filter by

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 08:58:25
            `memberOf`: WhereMixin
            `version`: 1.0
            `method_name`: where_timestamps
            * @xxx [04-25-2023 09:02:31]: documentation for where_timestamps
        '''
        # if self.model.get_column(column_name) is None:
        #     return None
        if start_timestamp is not None and end_timestamp is not None:
            self.between(column_name,start_timestamp,end_timestamp)
        if start_timestamp is None and end_timestamp is not None:
            self.less_than_equal(column_name,end_timestamp)
        if start_timestamp is not None and end_timestamp is None:
            self.greater_than_equal(column_name,start_timestamp)

    def match(self,columns:Union[Iterable[str],str],value:Union[Iterable[str],str])->_t.query_type:
        # if isinstance(value,(list)) is False:
        #     print(f"value:{value}")
        # @Mstep [] force the value to be a list of strings and remove any null values.
        value = c.arr.strip_list_nulls(c.arr.force_list(value))
        # @Mstep [] convert all values to strings.
        value = [str(x) for x in value]
        self.add_where(columns,value)
        return self

    def like(self,column:Union[str,list,_t.column_type],value:str):
        if isinstance(column,(list)) is False:
            column = c.arr.force_list(column)
        from volent.Column import Column as _column
        columns = []
        for col in column:
            if isinstance(col,(str)):
                co = self.model.get_column(col)
                if co is not None:
                    columns.append(co.name)
            if isinstance(col,(_column)):
                columns.append(col.name)
        for col in columns:
            self.add_where(col,value,"like")
        return self



    def or_(self,*or_children:_t.where_clause_type):
        # print(or_children)
        w = self.add_where(_t.undefined,[],"or")
        w.or_child_count = len(or_children)
        w.Query.gather_wheres(w)

    



def format_null(value):
    if value is None:
        return "NULL"
    return value





# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable
import colemen_utils as c
import volent.settings as _settings
import volent.settings.types as _t
from volent.data_types.TypeBase import TypeBase as _type_base

@dataclass
class TinyInt(_type_base):

    def __init__(self,data_length:int=None) -> None:
        super().__init__(data_length)
        self.min_data_length = 0
        self.max_data_length = 255
        self.sql_type_name = "TINYINT"
        self.python_data_type = (int)
        self.open_api_data_type = "integer"

        self._validate_data_length(data_length,_settings.control.varchar_default_length)


    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} : {self.data_length}>"






from volent.data_types.BigInt import BigInt as BigInt
from volent.data_types.EncodedPrimary import EncodedPrimary
from volent.data_types.Bool import Bool as Bool
from volent.data_types.Decimal import Decimal as Decimal
from volent.data_types.Integer import Integer as Integer
from volent.data_types.String import String as String
from volent.data_types.TinyInt import TinyInt as TinyInt
from volent.data_types.TypeBase import TypeBase as _TypeBase







# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
'''
    The query delete module.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 04-28-2023 07:24:17
    `name`: delete
    * @xxx [04-28-2023 07:24:33]: documentation for delete
'''



from dataclasses import dataclass
from typing import Iterable, Union


import colemen_utils as c


import volent.settings.types as _t
# import volent.settings as _settings
# from volent.Field import Field as _field
# from volent.Relationship import Relationship as _relationship
# from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint
from volent.query.Query import Query
from volent.query.WhereMixin import WhereMixin


@dataclass
class Delete(Query,WhereMixin):
    def __init__(self,model:_t.model_type) -> None:
        '''
            Create a delete query instance.

            ----------

            Arguments
            -------------------------
            `model` {model}
                The model instance to execute a delete operation on.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:24:53
            `memberOf`: Delete
            `version`: 1.0
            `method_name`: Delete
            * @xxx [04-28-2023 07:26:55]: documentation for Delete
        '''
        self._params = {}

        self.query_crud_type = "delete"
        super().__init__(model)

    @property
    def select_string(self)->str:
        '''
            Get the compiled select string for this query.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-09-2022 15:56:15
            `@memberOf`: SelectQuery
            `@property`: select_string
        '''
        if len(self._selects) == 0:
            value = "*"
        else:
            selects = []
            for sel in self._selects:
                value = None
                if sel['alias'] is not None:
                    value = f"{sel['column_name']} as {sel['alias']}"
                else:
                    value = f"{sel['column_name']}"
                if value is not None:
                    selects.append(value)
            value = ', '.join(selects)

        if self._count is True:
            value = f"COUNT({value})"
        if self._average is True:
            value = f"AVG({value})"
        if self.sum_ is True:
            value = f"SUM({value})"
        return value


    @property
    def query(self)->tuple:
        '''
            Generate this delete query's SQL.

            ----------

            Return {tuple}
            ----------------------
            A tuple containing the query string and a dictionary of query params.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:27:18
            `memberOf`: Delete
            `version`: 1.0
            `method_name`: query
            * @xxx [04-28-2023 07:33:51]: documentation for query
        '''

        value = f"DELETE {self.model.quoted_name} {self.where_string}"

        value = self._paginate_select_query(value)
        value = self._format_query_params(value,self._params)
        return value,self._params
        # return value,self._params

    def execute(self,foreign_key_checks:bool=True)->Union[bool,dict,int]:
        '''
            Execute the delete operation on the model.

            ----------

            Arguments
            -------------------------
            [`foreign_key_checks`=True] {bool}
                If False, the query will not perform foreign key checks before executing the deletion.
                Be careful, this can cause integrity issues and will NOT warn you about it.


            Return {bool}
            ----------------------
            True if execution was successful,False otherwise.


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:34:05
            `memberOf`: Delete
            `version`: 1.0
            `method_name`: execute
            * @xxx [04-28-2023 07:40:57]: documentation for execute
        '''

        sql,args= self.query

        if sql is False:
            return False
        # @Mstep [] execute the delete query.
        result = self.database.run(sql,args,foreign_key_checks=foreign_key_checks)
        return result




def format_null(value):
    if value is None:
        return "NULL"
    return value




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
import re
from typing import Iterable, OrderedDict, Pattern, Sequence, Sized, Any, TypeVar, Union
from collections import OrderedDict
from abc import ABC, abstractmethod


import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator


_T = TypeVar("_T")

class CreditCardNumber(Validator):

    field_name = None

    default_message = "Invalid credit card number"


    def __init__(
        self,
        error:str=None,
        ):
        '''
            A validator that passes if the value is a credit card number
            ----------

            Arguments
            -------------------------

            `error` {str}
                The custom error to use if this validator fails.
                Accepts `input`,`field_name`

            Return {any}
            ----------------------
            The test value is successful, raises ValidationError if it fails.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 03-27-2023 09:10:41
            `memberOf`: NoneOf
            `version`: 1.0
            `method_name`: NoneOf
            * @xxx [03-27-2023 09:13:12]: documentation for NoneOf
        '''
        self.regex = re.compile(r"(4[0-9]{12}(?:[0-9]{3})?$)|(^(?:5[1-5][0-9]{2}|222[1-9]|22[3-9][0-9]|2[3-6][0-9]{2}|27[01][0-9]|2720)[0-9]{12}$)|(3[47][0-9]{13})|(^3(?:0[0-5]|[68][0-9])[0-9]{11}$)|(^6(?:011|5[0-9]{2})[0-9]{12}$)|(^(?:2131|1800|35\d{3})\d{11})")
        self.error = error or self.default_message  # type: str


    def _format_error(self, value: Union[str,bytes]) -> str:
        args = {
            "input":value,
        }
        if self.field_name is not None:
            args["field_name"] = self.field_name

        return self.error.format(**args)


    def _repr_args(self) -> str:
        return f""

    def __call__(self,value:_T,field_name:str=None)->_T:
        self.field_name = field_name

        if luhn_checksum(value) != 0:
            raise ValidationError(self._format_error(value))
        else:
            return value








        # if re.search(self.regex,value) is None:
        # # if self.regex.match(value) is None:
        #     raise ValidationError(self._format_error(value))


        # return value



def luhn_checksum(card_number):
    def digits_of(n):
        return [int(d) for d in str(n)]
    digits = digits_of(card_number)
    odd_digits = digits[-1::-2]
    even_digits = digits[-2::-2]
    checksum = 0
    checksum += sum(odd_digits)
    for d in even_digits:
        checksum += sum(digits_of(d*2))
    return checksum % 10

# print('Valid') if luhn_checksum("4532015112830366")==0 else print('Invalid')



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


# from typing import Iterable, OrderedDict, Union
# from collections import OrderedDict
import re
import sqlite3
import sys
from typing import Iterable
import mysql.connector as _mysqlConnnector
from mysql.connector.errors import ProgrammingError
import traceback as _traceback
import colemen_utils as c
import volent.settings as _settings
import volent.settings.types as _t
import volent.exceptions as _e
from volent.mixins.MySQLGeneratorMixin import MySQLGeneratorMixin as _MySQLGeneratorMixin

_log = c.con.log


class DatabaseConnection(_MySQLGeneratorMixin):
    '''Class used as a mixin to allow database connections and the general CRUD procedures.'''
    creds = None
    connection = None
    cursor = None


    def __init__(self):
        self.creds = {}


    def get_databases_with_creds(self)->Iterable[str]:
        '''Retrieve a list of database names that have credentials in the environment.'''
        dbs = []
        for db in self.database_names:
            db_creds = c.build.get_environ(db,None)
            if db_creds is None:
                c.con.log(f"Failed to locate credentials for database: {db}","error")
                c.con.log(f"Add credentials to the environment in order for a connection to be established.","error")
                c.con.log('Example: tmp_creds = {    "database": "database_name",    "user": "username",    "password": "ps",    "host": "162.850.216.128"}',"error")
                continue
            else:
                dbs.append(db)
        return dbs



    def connect(self, db_creds:dict=None):
        '''
            Sets up the database connection with the initial settings.

            If the DB_CREDENTIALS are provided, it attempts to connect to a mysql database.

            ----------

            Keyword Arguments
            -----------------
            `DB_PATH` {string}
                The filepath to the sqlite database

            [`create`=True] {bool}
                If True and SQLite database does not exist yet, create the file.

            `DB_CREDENTIALS` {dict}
                The credentials to connect to the mysql database
                {
                    "user":"string",
                    "password":"string",
                    "host":"string",
                    "database":"string"
                }

            Return {bool}
            ----------
                True upon success, false otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-19-2021 08:04:18
            `version`: 1.0
        '''
        connect_success = False
        from volent.Database import Database as _Database
        if isinstance(self,_Database):
            db_creds = c.build.get_environ(self.name,None)
            if db_creds is None:
                c.con.log(f"Failed to locate credentials for database: {db}","error")
            else:
                if len(db_creds.keys()) == 1:
                    connect_success = self.__connect_to_sqlite(db_creds)
                    if connect_success:
                        c.con.log(f"Successfully connected to {db_creds['database']}","green")
                    return connect_success
                connect_success = self.__connect_to_my_sqldb(db_creds)
                if connect_success:
                    # _traceback.print_stack()
                    c.con.log(f"Successfully connected to {db_creds['database']}","green")
                    # connect_success = True
                    return True

        else:
            from volent.Volent import Volent as _bv
            if isinstance(self,_bv):
                for db in self.databases:
                    result = db.connect()
                    if result is False:
                        c.con.log(f"Failed to connect to {db.name}","warning")
                        connect_success = False
                        return False
                return True




        # if _settings.globe.flavor == "mysql":
        #     if db_creds is None:
        #         db_creds = self.credentials()
        #     if isinstance(db_creds,(dict)):
        #         connect_success = self.__connect_to_my_sqldb(db_creds)
        #         if connect_success:
        #             c.con.log(f"Successfully connected to {db_creds['database']}","green")

        return connect_success

    def __validate_db_credentials(self,creds:dict):
        '''
                Validates that all of the db_credentials are provided.

                ----------

                Return {bool}
                ----------------------
                True upon success, false otherwise.

                Meta
                ----------
                `author`: Colemen Atwood
                `created`: 04-19-2021 08:23:40
                `memberOf`: colemen_database
                `version`: 1.0
                `method_name`: __validate_db_credentials
        '''
        # if 'db_credentials' in self.data:
        error_array = []
        # creds = self.data['db_credentials']
        if 'user' not in creds:
            error_array.append('user is not provided in db_credentials')
        if 'password' not in creds:
            error_array.append('password is not provided in db_credentials')
        if 'host' not in creds:
            error_array.append('host is not provided in db_credentials')
        if 'database' not in creds:
            error_array.append('database is not provided in db_credentials')
        if len(error_array) == 0:
            # print("Successfully validated db_credentials")
            return True
        return False

        print("Credentials are needed to connect to the Mysql Database.")
        return False

    def __connect_to_my_sqldb(self,creds:dict):
        '''
            Attempts to connect to a mysql database.

            ----------

            Return {bool}
            ----------------------
            True upon success, false otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-19-2021 08:23:40
            `memberOf`: colemen_database
            `version`: 1.0
            `method_name`: __connect_to_my_sqldb
        '''

        connect_success = False
        if self.connection is not None:
            return True

        if self.__validate_db_credentials(creds) is True:
            _settings.globe.flavor = "mysql"
            self.connection = None
            try:

                self.connection = _mysqlConnnector.connect(
                    user=creds['user'],
                    password=creds['password'],
                    host=creds['host'],
                    database=creds['database']
                )
                self.cursor = self.connection.cursor(
                    buffered=True,
                    dictionary=True
                )

                if self.connection.is_connected():
                    # print("Successfully connected to mysql database")
                    connect_success = True

            except ProgrammingError as error:
                c.con.log(error.msg,"red")

        return connect_success


    def __connect_to_sqlite(self,creds:dict):
        '''
            Attempts to connect to a mysql database.

            ----------

            Return {bool}
            ----------------------
            True upon success, false otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-19-2021 08:23:40
            `memberOf`: colemen_database
            `version`: 1.0
            `method_name`: __connect_to_my_sqldb
        '''

        connect_success = False
        if self.connection is not None:
            return True
        _settings.globe.flavor = "sqlite"
        db_name = creds['database']
        file_name = f"{db_name}.db"
        self.connection = sqlite3.connect(file_name)
        self.cursor = self.connection.cursor()

        if c.file.exists(file_name):
            connect_success = True

        return connect_success



    @property
    def is_connected(self):
        if self.connection is not None:
            return True
        return False

    def run(self, sql:str, args=False,foreign_key_checks=True):
        '''
            Executes a query on the database.

            ----------

            Arguments
            -------------------------
            `sql` {string}
                The sql query to execute.

            `args` {list}
                A list of arguments to apply to the sql query

            [`foreign_key_checks`=True] {bool}
                If False, the database will be instructed to skip foreign_key_checks.

            Return {bool}
            ----------------------
            True upon success, false otherwise.

            if multiple statements are provided it will return True if ALL statements execute successfully.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-19-2021 10:07:54
            `memberOf`: colemen_database
            `version`: 1.0
            `method_name`: run
        '''

        # if foreign_key_checks is False:
            # self.execute_single_statement("SET FOREIGN_KEY_CHECKS = 0;")
            # sql = f"SET FOREIGN_KEY_CHECKS = 0;\n{sql}\nSET FOREIGN_KEY_CHECKS = 0;"
        statements = sql
        success = False
        # print(f"executing: {sql}")
        # if the sql is a string, split it into a list of statements
        if isinstance(sql, (str)):
            statements = _to_statement_list(sql)
        # print(f"statements:{statements}")
        if len(statements) > 1:
            # print(f"Multiple statements [{len(statements)}] found in sql.")
            success = True
            for statement in statements:
                res = self.execute_single_statement(statement, args,foreign_key_checks=foreign_key_checks)
                if res is False:
                    success = False


        if len(statements) == 1:
            return self.execute_single_statement(sql, args,foreign_key_checks=foreign_key_checks)
        return success


    def run_select(self,sql:str,args=False,default=None,limit=None,offset=None):
        '''
            Execute a select query on the database.

            ----------

            Arguments
            -------------------------
            `sql` {str}
                The Select query to execute.

            `args` {list,dict}
                The arguments to use in parameterized placeholders

            Keyword Arguments
            -------------------------
            [`default`=None] {any}
                The default value to return in the event of an error.

            [`limit`=None] {int}
                The maximum number of results to return

            [`offset`=None] {int}
                The index offset to apply to the query.

            Return {any}
            ----------------------
            The results of the query if successful, the default value otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-09-2022 11:12:16
            `memberOf`: MySQLDatabase
            `version`: 1.0
            `method_name`: run_select
            * @xxx [12-09-2022 11:15:35]: documentation for run_select
        '''
        # default = _obj.get_kwarg(['default'],None,None,**kwargs)
        # limit = _obj.get_kwarg(['limit'],None,(int),**kwargs)
        # offset = _obj.get_kwarg(['offset'],None,(int),**kwargs)

        sql = _paginate_select_query(sql,limit,offset)
        if isinstance(args,(dict)):
            sql = _format_query_params(sql,args)
        # print(f"sql:{sql}")
        if self.is_connected is False:
            c.con.log("Not Connected to a database dork.","warning")
        # if self.connect() is False:
            return default

        # _log(f"sql:{sql}","cyan")
        # _log(f"args:{args}","cyan")

        if self.run(sql,args):
            return self.fetchall()
        return default


    def run_select_internal(self,sql:str,select:_t.query_type,args=False,default=None,limit=None,offset=None):
        '''
            Execute a select query on the database.

            ----------

            Arguments
            -------------------------
            `sql` {str}
                The Select query to execute.

            `args` {list,dict}
                The arguments to use in parameterized placeholders

            Keyword Arguments
            -------------------------
            [`default`=None] {any}
                The default value to return in the event of an error.

            [`limit`=None] {int}
                The maximum number of results to return

            [`offset`=None] {int}
                The index offset to apply to the query.

            Return {any}
            ----------------------
            The results of the query if successful, the default value otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-09-2022 11:12:16
            `memberOf`: MySQLDatabase
            `version`: 1.0
            `method_name`: run_select
            * @xxx [12-09-2022 11:15:35]: documentation for run_select
        '''
        # default = _obj.get_kwarg(['default'],None,None,**kwargs)
        # limit = _obj.get_kwarg(['limit'],None,(int),**kwargs)
        # offset = _obj.get_kwarg(['offset'],None,(int),**kwargs)

        sql = _paginate_select_query(sql,limit,offset)
        if isinstance(args,(dict)):
            sql = _format_query_params(sql,args)
        # print(f"sql:{sql}")
        if self.is_connected is False:
            c.con.log("Not Connected to a database dork.","warning")
        # if self.connect() is False:
            return default

        # _log(f"sql:{sql}","cyan")
        # _log(f"args:{args}","cyan")


        self.cursor.row_factory = sqlite3.Row
        if self.run(sql,args):
            rows = self.cursor.fetchall()
            # print(rows)
            dicts = self._to_dict(rows)
            return dicts
            # contents = self.cursor.fetchall()
            # print(f"selects: ")
            # print(select._selects)
            # print(contents)
            # if len(select._selects) == 0:
            #     out = []
            #     for x in contents:
                    
                
            return self.fetchall()
        return default



    def close(self):
        '''
            Close the connection to the mySQL database.

            ----------

            Return {None}
            ----------------------
            None

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 06-05-2022 19:19:32
            `memberOf`: DatabaseManager
            `version`: 1.0
            `method_name`: close
            * @xxx [06-05-2022 19:20:10]: documentation for close
        '''

        self.connection.close()
        self.connection = None
        self.cursor = None


    # def __sqlite_execute_single_statement(self, sql:str, args=False,isTimeoutRetry=False,foreign_key_checks=True):


    def execute_single_statement(self, sql:str, args=False,isTimeoutRetry=False,foreign_key_checks=True):
        '''
            Executes a single SQL query on the database.

            ----------

            Arguments
            -------------------------
            `sql` {string}
                The SQL to be executed.

            `args` {list}
                A list of arguments for parameter substitution.

            Return {bool}
            ----------------------
            True upon success, false otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-09-2021 09:19:40
            `memberOf`: colemen_database
            `version`: 1.0
            `method_name`: execute_single_statement
        '''
        success = False
        if self.cursor is None or self.connection is None:
            print("Not connected to a database, aborting query.")
            if self.data['credentials'] is not None:
                self.connect()
        try:
            if foreign_key_checks is False:
                if _settings.globe.flavor == "mysql":
                    self.execute_single_statement("SET FOREIGN_KEY_CHECKS = 0;")
                if _settings.globe.flavor == "sqlite":
                    self.execute_single_statement("PRAGMA foreign_keys = OFF;")

            if args is False:
                # print(f"executing sql: ",sql)
                self.cursor.execute(sql)
            else:

                args = _sanitize_args(args)
                self.cursor.execute(sql, args)

                # print(f"result: ",result)

            self.connection.commit()
            # self.connection.
            success = True
            if foreign_key_checks is False:
                if _settings.globe.flavor == "mysql":
                    self.execute_single_statement("SET FOREIGN_KEY_CHECKS = 1;")
                if _settings.globe.flavor == "sqlite":
                    self.execute_single_statement("PRAGMA foreign_keys = ON;")


        except _mysqlConnnector.errors.IntegrityError as e:
            # _log(f"{_traceback.format_exc()}","error")
            # _log(f"SQL: {sql}","error")
            self._parse_interface_error(e)

        except _mysqlConnnector.errors.InterfaceError as e:
            if isTimeoutRetry is True:
                # _log(f"{_traceback.format_exc()}","error")
                # _log(f"SQL: {sql}","error")
                self._parse_interface_error(e)
            if isTimeoutRetry is False:
                self.cursor = None
                self.connection = None
                self.connect()
                return self.execute_single_statement(sql,args,True)

        except _mysqlConnnector.errors.DatabaseError as e:
            _log(f"{_traceback.format_exc()}","error")
            _log(f"SQL: {sql}","error")
            self._parse_interface_error(e)


        # except sqlite3.Warning as error:
        #     _log(f"Warning: {error}","error")
        #     _log(_traceback.format_exc(),"error")

        # except sqlite3.OperationalError as error:
        #     _log(f"Fatal Error: {error}","error")
        #     _log(_traceback.format_exc(),"error")

        except AttributeError:
            _log(f"{_traceback.format_exc()}\n","error")
            _log(f"{print(sys.exc_info()[2])}\n\n","error")
            _log(f"SQL: \033[38;2;(235);(64);(52)m{sql}")

        return success

    def _parse_interface_error(self,e):
        if "duplicate" in e.msg.lower():
            # Duplicate entry 'act_succesboobers' for key 'uq_hash_id'
            raise _e.DuplicateEntryError(e)
        # print(f"error:{e}")
        # print(f"e.args: {e.args}")
        # print(f"e.errno: {e.errno}")
        # print(f"e.msg: {e.msg}")
        # print(f"e.sqlstate: {e.sqlstate}")

    def run_from_list(self, query_list,**kwargs):
        '''
            Execute SQL statements from a list.

            ----------

            Arguments
            -------------------------
            `query_list` {list}
                A list of query statements to execute.

            Keyword Arguments
            -------------------------
            [`disable_restraints`=True] {bool}
                If True, temporarily disable foreign_key_checks while executing the queries

            Return {bool}
            ----------------------
            True upon success, false otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 06-05-2022 16:32:58
            `memberOf`: DatabaseManager
            `version`: 1.0
            `method_name`: run_from_list
            * @xxx [06-05-2022 16:36:56]: documentation for run_from_list
        '''

        disable_foreign_key_restraints = c.obj.get_kwarg(['disable key restraints','disable restraints'],True,(bool),**kwargs)
        # disable_foreign_key_restraints = True
        # if 'DISABLE_KEY_RESTRAINTS' in kwargs:
        #     if kwargs['DISABLE_KEY_RESTRAINTS'] is False:
        #         disable_foreign_key_restraints = False
        if disable_foreign_key_restraints is True:
            self.run("SET foreign_key_checks = 0;")

        success = True
        for idx,que in enumerate(query_list):
            print(f"{idx}/{len(query_list)}",end="\r",flush=True)
            success = self.run(que)
            if success is False:
                break

        if disable_foreign_key_restraints is True:
            self.run("SET foreign_key_checks = 1;")
        return success

    def run_multi(self, sql:str, args):
        sql = sql.replace(";", ";STATEMENT_END")
        statements = sql.split('STATEMENT_END')
        for s in statements:
            if len(s) > 0:
                # print(f"query: {s}")
                self.run(s, args)

    def fetchall(self):
        '''
            Executes the fetchall method on the database and converts the result to a dictionary.

            ----------


            Return {dict|list}
            ----------------------
            If there is more than one result, it returns a list of dicts.
            If there is only one result, it returns a single dictionary.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 06-02-2022 13:58:55
            `memberOf`: DatabaseManager
            `version`: 1.0
            `method_name`: fetchall
            * @xxx [06-02-2022 13:59:37]: documentation for fetchall
        '''
        # def dict_factory(cursor, row):
        #     d = {}
        #     for idx, col in enumerate(cursor.description):
        #         d[col[0]] = row[idx]
        #     return d
        # return dict_factory(self.cursor,self.cursor.fetchall())
        # colname = [ d[0] for d in query.description ]
        # result_list = [ dict(zip(colname, r)) for r in query.fetchall() ]
        
        result = self.cursor.fetchall()
        # print(f"fetchall::result : {result}")
        return result
        return self._to_dict(self.cursor.fetchall())

    def fetchone(self):
        """ DOCBLOCK {
                "class_name":"Database",
                "method_name":"fetchone",
                "author":"Colemen Atwood",
                "created": "04-19-2021 08:04:18",
                "version": "1.0",
                "description":"Executes the fetchone method on the database.",
                "returns":{
                    "type":"dict",
                    "description":"The result of the fetchone command"
                }
            }"""
        r = self.cursor.fetchone()
        return r

    def _to_dict(self, result):
        # print(f"_to_dict:resultType: {type(result)}")
        # print(f"_to_dict:result : {result}")
        if isinstance(result, (list)):
            new_data = []
            for row in result:
                tmp = {}
                # @Mstep [IF] if the row is a tuple
                # this means its probably the response of the sqlite last insert id.
                if isinstance(row,(tuple)):
                    if len(row) == 1:
                        row = row[0]
                    return row
                # print(f"row:{row.keys()}")
                for col in row.keys():
                    tmp[col] = row[col]
                new_data.append(tmp)
            return new_data
        # if isinstance(result, sqlite3.Row):
        #     new_data = {}
        #     for col in result.keys():
        #         new_data[col] = result[col]
        #     return new_data





# def _to_statement_list(sql):
#     sql = sql.replace(";", ";STATEMENT_END")
#     statements = sql.split('STATEMENT_END')
#     output = [x.strip() for x in statements if len(x.strip()) > 0]
#     return output


def _to_statement_list(sql):

    def format_triggers(sql):
        '''
        This will locate trigger create statements and modify the sql to ignore
        the internal colons.

        This is required in order for the statement splitting by colons to work,
        other wise is will split the trigger into two separate statements.
        '''
        sline_sql = sql.replace("\n","__NEW_LINE__")

        matches = re.findall(r'(CREATE\s*TRIGGER.*END;)',sline_sql)
        for match in matches:
            target = match.replace("__NEW_LINE__","\n")
            new = match.replace("END;","__TRIGGER_END__")
            new = new.replace(";", "__SINGLE_COLON__")
            new = new.replace("__TRIGGER_END__","END;")
            new = new.replace("__NEW_LINE__","\n")
            sql = sql.replace(target,new)
        return sql

    sql = format_triggers(sql)
    sql = sql.replace(";", ";STATEMENT_END")
    sql = sql.replace("__SINGLE_COLON__", ";")

    statements = sql.split('STATEMENT_END')
    # print(f"sql:{sql}")
    # _ = [print(f"{'-'*80}\n{x}") for x in statements]

    # print(f"statements:{statements}")
    # exit()
    # statements = []
    # if ";REAL_STATEMENT_END" in sql:
    #     st = sql.split('REAL_STATEMENT_END')
        # for s in st:


    output = [x.strip() for x in statements if len(x.strip()) > 0]
    return output

def _sanitize_args(args):
    if isinstance(args,(dict)):
        output = {}
        for k,v in args.items():
            output[k] = c.string.sanitize_quotes(v)
        return output
    if isinstance(args,(list)):
        output = []
        for v in args:
            output.append(c.string.sanitize_quotes(v))
        return output



def _paginate_select_query(sql:str,limit:int=None,offset:int=None)->str:
    '''
        Apply a limit and offset value to a select query statement.
        ----------

        Arguments
        -------------------------
        `sql` {str}
            The sql statement to modify

        [`limit`=None] {int}
            The limit to apply to the results.

        [`offset`=None] {int}
            The offset to apply to the results.


        Return {str}
        ----------------------
        The sql statement with a limit and offset value applied.

        If the limit/offset if invalid no pagination is added.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 12-09-2022 11:09:53
        `memberOf`: MySQLDatabase
        `version`: 1.0
        `method_name`: _paginate_select_query
        * @xxx [12-09-2022 11:11:52]: documentation for _paginate_select_query
    '''
    if limit is None and offset is None:
        return sql
    if isinstance(limit,(str)):
        limit = re.sub(r'[^0-9]',"",limit)
        if len(limit) == 0:
            return sql
        limit = int(limit)


    if isinstance(offset,(str)):
        offset = re.sub(r'[^0-9]',"",offset)
        if len(offset) == 0:
            offset = None
        else:
            offset = int(offset)

    if limit == 0:
        limit = 1

    if offset is not None:
        if offset < 1:
            offset = None

    sql = c.string.strip(sql,[";"],"right")
    sql = re.sub(r'limit\s*[0-9]*\s*(,|offset)\s*(:?[0-9\s]*)?',"",sql,re.MULTILINE | re.IGNORECASE)

    limit_string = f"LIMIT {limit}"
    offset_string = ""
    if offset is not None:
        offset_string = f"OFFSET {offset}"
    paginate = f"{limit_string} {offset_string}"
    sql = f"{sql} {paginate}"
    return sql


def _format_query_params(sql:str,args:dict)->str:
    '''
        Format an SQL query's parameters to use the python named template format.

        This will only replace matches that have a corresponding key in the args dictionary.

        Parameters can begin with a dollar sign or colon.


        SELECT * from blackholes WHERE hash_id=$hash_id

        SELECT * from blackholes WHERE hash_id=%(hash_id)s

        ----------

        Arguments
        -------------------------
        `sql` {str}
            The sql string to format.

        `args` {dict}
            The dictionary of parameter values.


        Return {str}
        ----------------------
        The sql statement with parameters replaced.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 12-09-2022 10:39:24
        `memberOf`: MySQLDatabase
        `version`: 1.0
        `method_name`: _format_query_params
        * @xxx [12-09-2022 10:43:23]: documentation for _format_query_params
    '''
    if isinstance(args,(dict)) is False:
        return sql
    # args = sorted(args.items(), key=lambda x: x[1], reverse=True)
    arg_keys = list(args.keys())
    arg_keys.sort(key=len, reverse=True)
    for k in arg_keys:
    # for k,v in args.items():
        sql = re.sub(fr'[$:]{k}',f"%({k})s",sql)

    # matches = re.findall(r'[$:]([a-z_0-9]*)',sql,re.IGNORECASE)
    # if isinstance(matches,(list)):
    #     for match in matches:
    #         if match in args:
    #             sql = sql.replace(f"${match}",f"%({match})s")

    return sql



